from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.policy.idm_policy import IDMPolicy
from metadrive.utils import setup_logger

if __name__ == "__main__":
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 1,
            "traffic_density": 0,
            "traffic_mode": "trigger",
            "start_seed": 22,
            # "_disable_detector_mask":True,
            # "debug_physics_world": True,
            "global_light": True,
            # "debug_static_world":True,
            "cull_scene": False,
            # "image_observation": True,
            # "controller": "joystick",
            # "manual_control": True,
            "use_render": False,
            "decision_repeat": 5,
            "need_inverse_traffic": True,
            "rgb_clip": True,
            "debug": False,
            "map": "yBY",
            # "agent_policy": IDMPolicy,
            "random_traffic": False,
            "random_lane_width": True,
            # "random_agent_model": True,
            "driving_reward": 1.0,
            "force_destroy": False,
            "vehicle_config": {
                "enable_reverse": False,
                # "image_source": "depth_camera",
                # "random_color": True
                # "show_lidar": True,
                # "spawn_lane_index":("1r1_0_", "1r1_1_", 0),
                # "destination":"2R1_3_",
                # "show_side_detector": True,
                # "show_lane_line_detector": True,
                # "side_detector": dict(num_lasers=2, distance=50),
                # "lane_line_detector": dict(num_lasers=2, distance=50),
                # # "show_line_to_dest": True,
                # "show_dest_mark": True
            },
        }
    )
    import time
    # [9.95036221 0.99503618]
    start = time.time()
    o, _ = env.reset()
    env.vehicle.set_velocity([1, 0.1], 10)
    # print(env.vehicle.speed)

    for s in range(1, 10000):
        o, r, tm, tc, info = env.step([1, 0.5])
        # print("heading: {} forward_direction: {}".format(env.vehicle.heading, env.vehicle.velocity_direction))

        # env.vehicle.set_velocity([1, 10], 10)
        # # print(env.vehicle.velocity)

        # if s % 100 == 0:
        #     env.close()
        #     env.reset()
        # info["fuel"] = env.vehicle.energy_consumption
        env.render(track_target_vehicle=True, mode="top_down")



import numpy as np

from metadrive.component.pgblock.create_pg_block_utils import CreateAdverseRoad, CreateRoadFrom, create_bend_straight
from metadrive.component.pgblock.pg_block import PGBlock
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class Curve(PGBlock):
    """
        2 - - - - - - - - - -
       / 3 - - - - - - - - - -
      / /
     / /
    0 1
    """
    ID = "C"
    SOCKET_NUM = 1
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.CURVE)

    def _try_plug_into_previous_block(self) -> bool:
        parameters = self.get_config()
        basic_lane = self.positive_basic_lane
        lane_num = self.positive_lane_num

        # part 1
        start_node = self.pre_block_socket.positive_road.end_node
        end_node = self.add_road_node()
        positive_road = Road(start_node, end_node)
        length = parameters[Parameter.length]
        direction = parameters[Parameter.dir]
        angle = parameters[Parameter.angle]
        radius = parameters[Parameter.radius]
        curve, straight = create_bend_straight(
            basic_lane, length, radius, np.deg2rad(angle), direction, basic_lane.width,
            (PGLineType.BROKEN, PGLineType.SIDE)
        )
        no_cross = CreateRoadFrom(
            curve,
            lane_num,
            positive_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        no_cross = CreateAdverseRoad(
            positive_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        # part 2
        start_node = end_node
        end_node = self.add_road_node()
        positive_road = Road(start_node, end_node)
        no_cross = CreateRoadFrom(
            straight,
            lane_num,
            positive_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        no_cross = CreateAdverseRoad(
            positive_road,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross

        # common properties
        self.add_sockets(self.create_socket_from_positive_road(positive_road))
        return no_cross



from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.parking_lot import ParkingLot
from metadrive.component.pgblock.std_intersection import StdInterSection
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.tests.vis_block.vis_block_base import TestBlock

if __name__ == "__main__":
    StdInterSection.EXIT_PART_LENGTH = 4
    test = TestBlock()
    from metadrive.engine.asset_loader import initialize_asset_loader

    initialize_asset_loader(test)

    global_network = NodeRoadNetwork()
    last = FirstPGBlock(global_network, 3, 1, test.render, test.world, 1)

    last = StdInterSection(1, last.get_socket(0), global_network, 1)
    last.construct_block(test.render, test.world, dict(radius=4))
    inter_1 = last

    last = ParkingLot(2, last.get_socket(1), global_network, 1)
    last.construct_block(test.render, test.world)

    last = StdInterSection(3, last.get_socket(0), global_network, 1)
    last.construct_block(test.render, test.world, dict(radius=4))
    inter_2 = last

    last = StdInterSection(4, last.get_socket(2), global_network, 1)
    last.construct_block(test.render, test.world, dict(radius=4))
    inter_3 = last

    last = ParkingLot(5, last.get_socket(2), global_network, 1)
    last.construct_block(test.render, test.world)

    inter_4 = StdInterSection(6, inter_1.get_socket(2), global_network, 1)
    inter_4.construct_block(test.render, test.world, dict(radius=4))

    # test.show_bounding_box(global_network)
    test.run()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division

from panda3d.core import CS_yup_right, CS_zup_right, invert, Vec3, Mat4, Vec4
from panda3d.core import SamplerState
from direct.stdpy.file import open

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.loader import RPLoader

from metadrive.render_pipeline.rpcore.util.shader_input_blocks import GroupedInputBlock


class CommonResources(RPObject):
    """ This class manages the loading and binding of commonly used resources,
    such as textures, models, but also shader inputs """
    def __init__(self, pipeline):
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._showbase = Globals.base
        self._ptas = {}
        self._load_fonts()
        self._load_textures()
        self._setup_inputs()

    def _load_fonts(self):
        """ Loads the default font used for rendering and assigns it to
        Globals.font for further usage """
        font = RPLoader.load_font("/$$rp/data/font/Roboto-Medium.ttf")
        font.set_pixels_per_unit(35)
        font.set_poly_margin(0.0)
        font.set_texture_margin(1)
        font.set_bg(Vec4(1, 1, 1, 0))
        font.set_fg(Vec4(1, 1, 1, 1))
        Globals.font = font

    def _setup_inputs(self):
        """ Creates commonly used shader inputs such as the current mvp and
        registers them to the stage manager so they can be used for rendering """

        self._input_ubo = GroupedInputBlock("MainSceneData")
        inputs = (
            ("camera_pos", "vec3"),
            ("view_proj_mat_no_jitter", "mat4"),
            ("last_view_proj_mat_no_jitter", "mat4"),
            ("last_inv_view_proj_mat_no_jitter", "mat4"),
            ("view_mat_z_up", "mat4"),
            ("proj_mat", "mat4"),
            ("inv_proj_mat", "mat4"),
            ("view_mat_billboard", "mat4"),
            ("frame_delta", "float"),
            ("smooth_frame_delta", "float"),
            ("frame_time", "float"),
            ("current_film_offset", "vec2"),
            ("frame_index", "int"),
            ("screen_size", "ivec2"),
            ("native_screen_size", "ivec2"),
            ("lc_tile_count", "ivec2"),
            ("ws_frustum_directions", "mat4"),
            ("vs_frustum_directions", "mat4"),
        )
        for name, ipt_type in inputs:
            self._input_ubo.register_pta(name, ipt_type)

        self._pipeline.stage_mgr.input_blocks.append(self._input_ubo)

        # Main camera and main render have to be regular inputs, since they are
        # used in the shaders by that name.
        self._pipeline.stage_mgr.inputs["mainCam"] = self._showbase.cam
        self._pipeline.stage_mgr.inputs["mainRender"] = self._showbase.render

        # Set the correct frame rate interval
        Globals.clock.set_average_frame_rate_interval(3.0)

        # Set initial value for view_proj_mat_no_jitter
        view_mat = Globals.render.get_transform(self._showbase.cam).get_mat()
        proj_mat = Mat4(self._showbase.camLens.get_projection_mat())
        proj_mat.set_cell(1, 0, 0.0)
        proj_mat.set_cell(1, 1, 0.0)
        self._input_ubo.update_input("view_proj_mat_no_jitter", view_mat * proj_mat)

    def write_config(self):
        """ Generates the shader configuration for the common inputs """
        content = self._input_ubo.generate_shader_code()
        try:
            # Try to write the temporary file
            with open("/$$rptemp/$$main_scene_data.inc.glsl", "w") as handle:
                handle.write(content)
        except IOError as msg:
            self.error("Failed to write common resources shader configuration!", msg)

    def _load_textures(self):
        """ Loads commonly used textures and makes them available via the
        stage manager """
        self._load_environment_cubemap()
        self._load_prefilter_brdf()
        self._load_skydome()

    def _load_environment_cubemap(self):
        """ Loads the default cubemap used for the environment, which is used
        when no other environment data is available """
        envmap = RPLoader.load_cube_map("/$$rp/data/default_cubemap/cubemap.txo", read_mipmaps=True)
        envmap.set_minfilter(SamplerState.FT_linear_mipmap_linear)
        # envmap.set_format(Image.F_rgba16)
        envmap.set_magfilter(SamplerState.FT_linear)
        envmap.set_wrap_u(SamplerState.WM_repeat)
        envmap.set_wrap_v(SamplerState.WM_repeat)
        envmap.set_wrap_w(SamplerState.WM_repeat)
        self._pipeline.stage_mgr.inputs["DefaultEnvmap"] = envmap

    def _load_prefilter_brdf(self):
        """ Loads the prefiltered brdf """
        luts = [
            {
                "src": "slices/env_brdf_#.png",
                "input": "PrefilteredBRDF"
            },
            {
                "src": "slices_metal/env_brdf.png",
                "input": "PrefilteredMetalBRDF"
            },
            {
                "src": "slices_coat/env_brdf.png",
                "input": "PrefilteredCoatBRDF"
            },
        ]

        for config in luts:
            loader_method = RPLoader.load_texture
            if "#" in config["src"]:
                loader_method = RPLoader.load_3d_texture

            brdf_tex = loader_method("/$$rp/data/environment_brdf/{}".format(config["src"]))
            brdf_tex.set_minfilter(SamplerState.FT_linear)
            brdf_tex.set_magfilter(SamplerState.FT_linear)
            brdf_tex.set_wrap_u(SamplerState.WM_clamp)
            brdf_tex.set_wrap_v(SamplerState.WM_clamp)
            brdf_tex.set_wrap_w(SamplerState.WM_clamp)
            brdf_tex.set_anisotropic_degree(0)
            self._pipeline.stage_mgr.inputs[config["input"]] = brdf_tex

    def _load_skydome(self):
        """ Loads the skydome """
        skydome = RPLoader.load_texture("/$$rp/data/builtin_models/skybox/skybox.txo")
        skydome.set_wrap_u(SamplerState.WM_clamp)
        skydome.set_wrap_v(SamplerState.WM_clamp)
        self._pipeline.stage_mgr.inputs["DefaultSkydome"] = skydome

    def load_default_skybox(self):
        skybox = RPLoader.load_model("/$$rp/data/builtin_models/skybox/skybox.bam")
        return skybox

    def update(self):
        """ Updates the commonly used resources, mostly the shader inputs """
        update = self._input_ubo.update_input

        # Get the current transform matrix of the camera
        view_mat = Globals.render.get_transform(self._showbase.cam).get_mat()

        # Compute the view matrix, but with a z-up coordinate system
        zup_conversion = Mat4.convert_mat(CS_zup_right, CS_yup_right)
        update("view_mat_z_up", view_mat * zup_conversion)

        # Compute the view matrix without the camera rotation
        view_mat_billboard = Mat4(view_mat)
        view_mat_billboard.set_row(0, Vec3(1, 0, 0))
        view_mat_billboard.set_row(1, Vec3(0, 1, 0))
        view_mat_billboard.set_row(2, Vec3(0, 0, 1))
        update("view_mat_billboard", view_mat_billboard)

        update("camera_pos", self._showbase.camera.get_pos(Globals.render))

        # Compute last view projection mat
        curr_vp = self._input_ubo.get_input("view_proj_mat_no_jitter")
        update("last_view_proj_mat_no_jitter", curr_vp)
        curr_vp = Mat4(curr_vp)
        curr_vp.invert_in_place()
        curr_inv_vp = curr_vp
        update("last_inv_view_proj_mat_no_jitter", curr_inv_vp)

        proj_mat = Mat4(self._showbase.camLens.get_projection_mat())

        # Set the projection matrix as an input, but convert it to the correct
        # coordinate system before.
        proj_mat_zup = Mat4.convert_mat(CS_yup_right, CS_zup_right) * proj_mat
        update("proj_mat", proj_mat_zup)

        # Set the inverse projection matrix
        update("inv_proj_mat", invert(proj_mat_zup))

        # Remove jitter and set the new view projection mat
        proj_mat.set_cell(1, 0, 0.0)
        proj_mat.set_cell(1, 1, 0.0)
        update("view_proj_mat_no_jitter", view_mat * proj_mat)

        # Store the frame delta
        update("frame_delta", Globals.clock.get_dt())
        update("smooth_frame_delta", 1.0 / max(1e-5, Globals.clock.get_average_frame_rate()))
        update("frame_time", Globals.clock.get_frame_time())

        # Store the current film offset, we use this to compute the pixel-perfect
        # velocity, which is otherwise not possible. Usually this is always 0
        # except when SMAA and reprojection is enabled
        update("current_film_offset", self._showbase.camLens.get_film_offset())
        update("frame_index", Globals.clock.get_frame_count())

        # Compute frustum corners in the order BL, BR, TL, TR
        ws_frustum_directions = Mat4()
        vs_frustum_directions = Mat4()
        inv_proj_mat = Globals.base.camLens.get_projection_mat_inv()
        view_mat_inv = Mat4(view_mat)
        view_mat_inv.invert_in_place()

        for i, point in enumerate(((-1, -1), (1, -1), (-1, 1), (1, 1))):
            result = inv_proj_mat.xform(Vec4(point[0], point[1], 1.0, 1.0))
            vs_dir = (zup_conversion.xform(result)).xyz.normalized()
            vs_frustum_directions.set_row(i, Vec4(vs_dir, 1))
            ws_dir = view_mat_inv.xform(Vec4(result.xyz, 0))
            ws_frustum_directions.set_row(i, ws_dir)

        update("vs_frustum_directions", vs_frustum_directions)
        update("ws_frustum_directions", ws_frustum_directions)

        update("screen_size", Globals.resolution)
        update("native_screen_size", Globals.native_resolution)
        update("lc_tile_count", self._pipeline.light_mgr.num_tiles)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# This script downloads and updates the module builder.

# pylint: skip-file

if __name__ == "__main__":
    import os
    import sys
    from os.path import dirname, realpath, isfile

    # Files which are skipped from the downloaded zip
    files_to_ignore = "__init__.py LICENSE README.md config.ini config_module".split()

    current_dir = dirname(realpath(__file__))
    os.chdir(current_dir)

    # Import thirdparty packages
    sys.path.insert(0, "../../rpcore/util")
    sys.path.insert(0, "../../")

    # Download the module
    from submodule_downloader import download_submodule
    download_submodule("tobspr", "P3DModuleBuilder", current_dir, files_to_ignore)

    # Make an init script so we can import the code
    with open("scripts/__init__.py", "w") as handle:
        handle.write("")

    # Update the gitignore using the suggested version from the module builder
    if isfile(".gitignore"):
        os.remove(".gitignore")
    os.rename("prefab.gitignore", ".gitignore")



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

from panda3d.core import Notify, LineStream, Vec3

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.rpobject import RPObject

from metadrive.render_pipeline.rpcore.gui.text import Text


class ErrorMessageDisplay(RPObject):
    """ This is a gui element which listens to the panda output stream
    and shows errors """
    def __init__(self):
        RPObject.__init__(self)
        self._num_errors = 0
        self._error_node = Globals.base.pixel2d.attach_new_node("ErrorDisplay")
        self._error_node.set_z(-180)
        self._notify_stream = None

    def _init_notify(self):
        """ Internal method to init the stream to catch all notify messages """
        self._notify_stream = LineStream()
        Notify.ptr().set_ostream_ptr(self._notify_stream, False)

    def update(self):
        """ Updates the error display, fetching all new messages from the notify
        stream """

        if not self._notify_stream:
            self._init_notify()

        while self._notify_stream.is_text_available():
            line = self._notify_stream.get_line().strip()
            if "warning" in line:
                RPObject.global_warn("Panda3D", line)
                # self.add_warning(line)
            elif "error" in line:
                RPObject.global_error("Panda3D", line)
                self.add_error(line)
            else:
                RPObject.global_debug("Panda3D", line)

    def add_error(self, msg):
        """ Adds a new error message """
        self.add_text(msg, Vec3(1, 0.2, 0.2))

    def add_warning(self, msg):
        """ Adds a new warning message """
        self.add_text(msg, Vec3(1, 1, 0.2))

    def add_text(self, text, color):
        """ Internal method to add a new text to the output """
        Text(
            x=Globals.native_resolution.x - 30,
            y=self._num_errors * 23,
            align="right",
            text=text,
            size=12,
            parent=self._error_node,
            color=color
        )

        self._num_errors += 1

        if self._num_errors > 30:
            self.clear_messages()
            self.add_error("Error count exceeded. Cleared errors ..")

    def clear_messages(self):
        """ Clears all messages / removes them """
        self._error_node.node().remove_all_children()
        self._num_errors = 0

    def show(self):
        self._error_node.show()

    def hide(self):
        self._error_node.hide()



from panda3d.core import PNMImage, Texture, LVecBase3d, NodePath, Shader, LVecBase3i
from panda3d.core import ShaderAttrib, LVecBase2i, Vec2

from metadrive.render_pipeline.rpcore.globals import Globals

import math


class GPUFFT:
    """ This is a collection of compute shaders to generate the inverse
    fft efficiently on the gpu, with butterfly FFT and precomputed weights """
    def __init__(self, size, source_tex, normalization_factor):
        """ Creates a new fft instance. The source texture has to specified
        from the begining, as the shaderAttributes are pregenerated for
        performance reasons """

        self.size = size
        self.log2_size = int(math.log(size, 2))
        self.normalization_factor = normalization_factor

        # Create a ping and a pong texture, because we can't write to the
        # same texture while reading to it (that would lead to unexpected
        # behaviour, we could solve that by using an appropriate thread size,
        # but it works fine so far)
        self.ping_texture = Texture("FFTPing")
        self.ping_texture.setup_2d_texture(self.size, self.size, Texture.TFloat, Texture.FRgba32)
        self.pong_texture = Texture("FFTPong")
        self.pong_texture.setup_2d_texture(self.size, self.size, Texture.TFloat, Texture.FRgba32)
        self.source_tex = source_tex

        for tex in [self.ping_texture, self.pong_texture, source_tex]:
            tex.set_minfilter(Texture.FTNearest)
            tex.set_magfilter(Texture.FTNearest)
            tex.set_wrap_u(Texture.WMClamp)
            tex.set_wrap_v(Texture.WMClamp)

        # Pregenerate weights & indices for the shaders
        self._compute_weighting()

        # Pre generate the shaders, we have 2 passes: Horizontal and Vertical
        # which both execute log2(N) times with varying radii
        self.horizontal_fft_shader = Shader.load_compute(
            Shader.SLGLSL, "/$$rp/rpcore/water/shader/horizontal_fft.compute"
        )
        self.horizontal_fft = NodePath("HorizontalFFT")
        self.horizontal_fft.set_shader(self.horizontal_fft_shader)
        self.horizontal_fft.set_shader_input("precomputedWeights", self.weights_lookup_tex)
        self.horizontal_fft.set_shader_input("N", LVecBase2i(self.size))

        self.vertical_fft_shader = Shader.load_compute(Shader.SLGLSL, "/$$rp/rpcore/water/shader/vertical_fft.compute")
        self.vertical_fft = NodePath("VerticalFFT")
        self.vertical_fft.set_shader(self.vertical_fft_shader)
        self.vertical_fft.set_shader_input("precomputedWeights", self.weights_lookup_tex)
        self.vertical_fft.set_shader_input("N", LVecBase2i(self.size))

        # Create a texture where the result is stored
        self.result_texture = Texture("Result")
        self.result_texture.setup2dTexture(self.size, self.size, Texture.TFloat, Texture.FRgba16)
        self.result_texture.set_minfilter(Texture.FTLinear)
        self.result_texture.set_magfilter(Texture.FTLinear)

        # Prepare the shader attributes, so we don't have to regenerate them
        # every frame -> That is VERY slow (3ms per fft instance)
        self._prepare_attributes()

    def get_result_texture(self):
        """ Returns the result texture, only contains valid data after execute
        was called at least once """
        return self.result_texture

    def _generate_indices(self, storage_a, storage_b):
        """ This method generates the precompute indices, see
        http://cnx.org/content/m12012/latest/image1.png """
        num_iter = self.size
        offset = 1
        step = 0
        for i in range(self.log2_size):
            num_iter = num_iter >> 1
            step = offset
            for j in range(self.size):
                goLeft = (j // step) % 2 == 1
                index_a, index_b = 0, 0
                if goLeft:
                    index_a, index_b = j - step, j
                else:
                    index_a, index_b = j, j + step

                storage_a[i][j] = index_a
                storage_b[i][j] = index_b
            offset = offset << 1

    def _generate_weights(self, storage):
        """ This method generates the precomputed weights """

        # Using a custom pi variable should force the calculations to use
        # high precision (I hope so)
        pi = 3.141592653589793238462643383
        num_iter = self.size // 2
        num_k = 1
        resolution_float = float(self.size)
        for i in range(self.log2_size):
            start = 0
            end = 2 * num_k
            for b in range(num_iter):
                K = 0
                for k in range(start, end, 2):
                    fK = float(K)
                    f_num_iter = float(num_iter)
                    weight_a = Vec2(
                        math.cos(2.0 * pi * fK * f_num_iter / resolution_float),
                        -math.sin(2.0 * pi * fK * f_num_iter / resolution_float)
                    )
                    weight_b = Vec2(
                        -math.cos(2.0 * pi * fK * f_num_iter / resolution_float),
                        math.sin(2.0 * pi * fK * f_num_iter / resolution_float)
                    )
                    storage[i][k // 2] = weight_a
                    storage[i][k // 2 + num_k] = weight_b
                    K += 1
                start += 4 * num_k
                end = start + 2 * num_k

            num_iter = num_iter >> 1
            num_k = num_k << 1

    def _reverse_row(self, indices):
        """ Reverses the bits in the given row. This is required for inverse
        fft (actually we perform a normal fft, but reversing the bits gives
        us an inverse fft) """
        mask = 0x1
        for j in range(self.size):
            val = 0x0
            temp = int(indices[j])  # Int is required, for making a copy
            for i in range(self.log2_size):
                t = mask & temp
                val = (val << 1) | t
                temp = temp >> 1
            indices[j] = val

    def _compute_weighting(self):
        """ Precomputes the weights & indices, and stores them in a texture """
        indices_a = [[0 for i in range(self.size)] for k in range(self.log2_size)]
        indices_b = [[0 for i in range(self.size)] for k in range(self.log2_size)]
        weights = [[Vec2(0.0) for i in range(self.size)] for k in range(self.log2_size)]

        # Pre-Generating indices ..
        self._generate_indices(indices_a, indices_b)
        self._reverse_row(indices_a[0])
        self._reverse_row(indices_b[0])

        # Pre-Generating weights .."
        self._generate_weights(weights)

        # Create storage for the weights & indices
        self.weights_lookup = PNMImage(self.size, self.log2_size, 4)
        self.weights_lookup.setMaxval((2**16) - 1)
        self.weights_lookup.fill(0.0)

        # Populate storage
        for x in range(self.size):
            for y in range(self.log2_size):
                index_a = indices_a[y][x]
                index_b = indices_b[y][x]
                weight = weights[y][x]

                self.weights_lookup.set_red(x, y, index_a / float(self.size))
                self.weights_lookup.set_green(x, y, index_b / float(self.size))
                self.weights_lookup.set_blue(x, y, weight.x * 0.5 + 0.5)
                self.weights_lookup.set_alpha(x, y, weight.y * 0.5 + 0.5)

        # Convert storage to texture so we can use it in a shader
        self.weights_lookup_tex = Texture("Weights Lookup")
        self.weights_lookup_tex.load(self.weights_lookup)
        self.weights_lookup_tex.set_format(Texture.FRgba16)
        self.weights_lookup_tex.set_minfilter(Texture.FTNearest)
        self.weights_lookup_tex.set_magfilter(Texture.FTNearest)
        self.weights_lookup_tex.set_wrap_u(Texture.WMClamp)
        self.weights_lookup_tex.set_wrap_v(Texture.WMClamp)

    def _prepare_attributes(self):
        """ Prepares all shaderAttributes, so that we have a list of
        ShaderAttributes we can simply walk through in the update method,
        that is MUCH faster than using set_shader_input, as each call to
        set_shader_input forces the generation of a new ShaderAttrib """
        self.attributes = []
        textures = [self.ping_texture, self.pong_texture]

        current_index = 0
        firstPass = True

        # Horizontal
        for step in range(self.log2_size):
            source = textures[current_index]
            dest = textures[1 - current_index]

            if firstPass:
                source = self.source_tex
                firstPass = False

            index = self.log2_size - step - 1
            self.horizontal_fft.set_shader_input("source", source)
            self.horizontal_fft.set_shader_input("dest", dest)
            self.horizontal_fft.set_shader_input("butterflyIndex", LVecBase2i(index))
            self._queue_shader(self.horizontal_fft)
            current_index = 1 - current_index

        # Vertical
        for step in range(self.log2_size):
            source = textures[current_index]
            dest = textures[1 - current_index]
            is_last_pass = step == self.log2_size - 1
            if is_last_pass:
                dest = self.result_texture
            index = self.log2_size - step - 1
            self.vertical_fft.set_shader_input("source", source)
            self.vertical_fft.set_shader_input("dest", dest)
            self.vertical_fft.set_shader_input("isLastPass", is_last_pass)
            self.vertical_fft.set_shader_input("normalizationFactor", self.normalization_factor)
            self.vertical_fft.set_shader_input("butterflyIndex", LVecBase2i(index))
            self._queue_shader(self.vertical_fft)

            current_index = 1 - current_index

    def execute(self):
        """ Executes the inverse fft once """
        for attr in self.attributes:
            self._execute_shader(attr)

    def _queue_shader(self, node):
        """ Internal method to fetch the ShaderAttrib of a node and store it
        in the update queue """
        sattr = node.getAttrib(ShaderAttrib)
        self.attributes.append(sattr)

    def _execute_shader(self, sattr):
        """ Internal method to execute a shader by a given ShaderAttrib """
        Globals.base.graphicsEngine.dispatch_compute(
            (self.size // 16, self.size // 16, 1), sattr, Globals.base.win.get_gsg()
        )



from metadrive.component.pgblock.intersection import InterSection
from metadrive.component.pg_space import Parameter


class StdInterSection(InterSection):
    def _try_plug_into_previous_block(self) -> bool:
        self._config[Parameter.change_lane_num] = 0
        success = super(StdInterSection, self)._try_plug_into_previous_block()
        return success



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import io
import os
import struct

from . import spec


class GLTFBuffer(object):
    def __init__(self, filepath):
        self._filepath = filepath
        self._channels = []
        self._metadata = []

    def add_channel(self, metadata):
        self._channels.append(io.BytesIO())
        self._metadata.append(metadata)
        self._metadata[-1]['bufferView'] = len(self._metadata) - 1
        self._metadata[-1]['count'] = 0
        return self._metadata[-1]

    def write(self, channel_id, *values):
        size = {
            'SCALAR': 1,
            'VEC2': 2,
            'VEC3': 3,
            'VEC4': 4,
            'MAT4': 4 * 4,
        }[self._metadata[channel_id]['type']]
        assert size == len(values)

        if 'min' not in self._metadata[channel_id]:
            self._metadata[channel_id]['min'] = [None] * size

        if 'max' not in self._metadata[channel_id]:
            self._metadata[channel_id]['max'] = [None] * size

        for i, value in enumerate(values):
            if self._metadata[channel_id]['min'][i] is None:
                self._metadata[channel_id]['min'][i] = value
            else:
                self._metadata[channel_id]['min'][i] = min(self._metadata[channel_id]['min'][i], value)

            if self._metadata[channel_id]['max'][i] is None:
                self._metadata[channel_id]['max'][i] = value
            else:
                self._metadata[channel_id]['max'][i] = max(self._metadata[channel_id]['max'][i], value)

        type_ = {
            spec.TYPE_UNSIGNED_BYTE: 'B',
            spec.TYPE_UNSIGNED_SHORT: 'H',
            spec.TYPE_UNSIGNED_INT: 'I',
            spec.TYPE_FLOAT: 'f',
        }[self._metadata[channel_id]['componentType']]

        format_ = '<{}'.format(type_ * size)

        self._channels[channel_id].write(struct.pack(format_, *values))
        self._metadata[channel_id]['count'] += 1

    def write_raw(self, channel_id, data):
        self._channels[channel_id].write(data)
        self._metadata[channel_id]['count'] += len(data)

    def count(self, channel_id):
        return self._metadata[channel_id]['count']

    def export(self, parent_node, filepath=None):
        offset = 0
        data = bytearray()
        uri = 'nothing.bin'

        # accessors + buffer views
        for i in range(len(self._channels)):
            channel = self._channels[i]
            metadata = self._metadata[i]
            extras = metadata.get('extras') or {}
            parent_node['accessors'].append(metadata)

            part = channel.getbuffer()
            view = {
                'buffer': len(parent_node['buffers']),
                'byteLength': len(part),
                'byteOffset': offset,
                'extras': extras,
            }
            parent_node['bufferViews'].append(view)

            offset += len(part)
            data += part

        # embedded images + buffer views
        for gltf_image in parent_node.get('images', []):
            extras = gltf_image.get('extras') or {}

            part = None

            if 'uri' in extras:
                tfilepath = os.path.join(os.path.dirname(self._filepath), extras['uri'])
                with open(tfilepath, 'rb') as f:
                    part = f.read()

            elif 'data' in extras:
                part = extras.pop('data')

            if not part:
                continue

            view = {
                'buffer': len(parent_node['buffers']),
                'byteLength': len(part),
                'byteOffset': offset,
                'extras': extras,
            }
            parent_node['bufferViews'].append(view)
            gltf_image['bufferView'] = len(parent_node['bufferViews']) - 1

            offset += len(part)
            data += part

        if filepath:
            buffer_fp = filepath.replace('.gltf', '.bin')
            with open(buffer_fp, 'wb') as f:
                f.write(data)

            uri = os.path.relpath(buffer_fp, os.path.dirname(self._filepath))

        if len(data):
            gltf_buffer = {
                'byteLength': len(data),
            }
            if filepath:
                gltf_buffer['uri'] = uri
            parent_node['buffers'].append(gltf_buffer)

        return data



import logging

import numpy as np

from metadrive.component.map.base_map import BaseMap
from metadrive.component.road_network.edge_road_network import EdgeRoadNetwork
from metadrive.component.scenario_block.scenario_block import ScenarioBlock
from metadrive.engine.asset_loader import AssetLoader
from metadrive.type import MetaDriveType
from metadrive.scenario.scenario_description import ScenarioDescription
from metadrive.utils.math import resample_polyline, get_polyline_length


class ScenarioMap(BaseMap):
    def __init__(self, map_index, random_seed=None):
        self.map_index = map_index
        self.need_lane_localization = self.engine.global_config["need_lane_localization"]
        super(ScenarioMap, self).__init__(dict(id=self.map_index), random_seed=random_seed)

    def show_coordinates(self):
        lanes = [lane_info.lane for lane_info in self.road_network.graph.values()]
        self.engine.show_lane_coordinates(lanes)

    def _generate(self):
        block = ScenarioBlock(
            block_index=0,
            global_network=self.road_network,
            random_seed=0,
            map_index=self.map_index,
            need_lane_localization=self.need_lane_localization
        )
        block.construct_block(self.engine.worldNP, self.engine.physics_world, attach_to_world=True)
        self.blocks.append(block)

    def play(self):
        # For debug
        for b in self.blocks:
            b._create_in_world(skip=True)
            b.attach_to_world(self.engine.worldNP, self.engine.physics_world)
            b.detach_from_world(self.engine.physics_world)

    @property
    def road_network_type(self):
        return EdgeRoadNetwork

    def destroy(self):
        self.map_index = None
        super(ScenarioMap, self).destroy()

    def __del__(self):
        # self.destroy()
        logging.debug("Map is Released")
        # print("[ScenarioMap] Map is Released")

    def get_boundary_line_vector(self, interval):
        ret = {}
        for lane_id, data in self.blocks[-1].map_data.items():
            type = data.get("type", None)
            map_feat_id = str(lane_id)
            if MetaDriveType.is_road_line(type):
                if len(data[ScenarioDescription.POLYLINE]) <= 1:
                    continue
                line = np.asarray(data[ScenarioDescription.POLYLINE])[..., :2]
                length = get_polyline_length(line)
                resampled = resample_polyline(line, interval) if length > interval * 2 else line
                if MetaDriveType.is_broken_line(type):
                    ret[map_feat_id] = {
                        "type": MetaDriveType.LINE_BROKEN_SINGLE_YELLOW
                        if MetaDriveType.is_yellow_line(type) else MetaDriveType.LINE_BROKEN_SINGLE_WHITE,
                        "polyline": resampled
                    }
                else:
                    ret[map_feat_id] = {
                        "polyline": resampled,
                        "type": MetaDriveType.LINE_SOLID_SINGLE_YELLOW
                        if MetaDriveType.is_yellow_line(type) else MetaDriveType.LINE_SOLID_SINGLE_WHITE
                    }
            elif MetaDriveType.is_road_edge(type):
                line = np.asarray(data[ScenarioDescription.POLYLINE])[..., :2]
                length = get_polyline_length(line)
                resampled = resample_polyline(line, interval) if length > interval * 2 else line
                ret[map_feat_id] = {"polyline": resampled, "type": MetaDriveType.BOUNDARY_LINE}
            elif type == MetaDriveType.LANE_SURFACE_STREET:
                continue
            # else:
            # # for debug
            #     raise ValueError
        return ret

    # def get_map_features(self, interval=2):
    #     """
    #
    #     """
    #     map_features = super(ScenarioMap, self).get_map_features(interval=interval)
    #
    #     # Adding the information stored in original data to here
    #     original_map_features = self.engine.data_manager.get_scenario(self.map_index)["map_features"]
    #
    #     for map_feat_key, old_map_feat in original_map_features.items():
    #
    #         if map_feat_key not in map_features:
    #             # Discard the data for those map features that are not reconstructed by MetaDrive.
    #             pass
    #
    #             # old_map_feat = copy.deepcopy(old_map_feat)
    #             # old_map_feat["valid"] = False
    #             #
    #             # map_features[map_feat_key] = old_map_feat
    #             #
    #             # if "polyline" in old_map_feat:
    #             #     map_features[map_feat_key]["polyline"] = convert_polyline_to_metadrive(
    #             #         old_map_feat["polyline"], coordinate_transform=self.coordinate_transform
    #             #     )
    #             #
    #             # if "position" in old_map_feat:
    #             #     if self.coordinate_transform:
    #             #         map_features[map_feat_key]["position"] = waymo_to_metadrive_vector(old_map_feat["position"])
    #             #     else:
    #             #         map_features[map_feat_key]["position"] = old_map_feat["position"]
    #
    #         else:
    #             # This map features are in both original data and current data.
    #             # We will check if in original data this map features contains some useful metadata.
    #             # If so, copied it to the new data.
    #             if "speed_limit_kmh" in old_map_feat:
    #                 map_features[map_feat_key]["speed_limit_kmh"] = old_map_feat["speed_limit_kmh"]
    #             if "speed_limit_mph" in old_map_feat:
    #                 map_features[map_feat_key]["speed_limit_mph"] = old_map_feat["speed_limit_mph"]
    #
    #     return map_features


if __name__ == "__main__":
    from metadrive.engine.engine_utils import initialize_engine
    from metadrive.envs.scenario_env import ScenarioEnv
    from metadrive.manager.scenario_data_manager import ScenarioDataManager

    # # touch these items so that pickle can work

    # file_path = AssetLoader.file_path("waymo", "0.pkl", return_raw_style=False)
    # # file_path = "/home/shady/Downloads/test_processed/60.pkl"
    # data = read_scenario_data(file_path)

    default_config = ScenarioEnv.default_config()
    default_config["use_render"] = True
    default_config["debug"] = True
    default_config["debug_static_world"] = True
    default_config["data_directory"] = AssetLoader.file_path("waymo", return_raw_style=False)
    # default_config["data_directory"] = AssetLoader.file_path("nuscenes", return_raw_style=False)
    # default_config["data_directory"] = "/home/shady/Downloads/test_processed"
    default_config["num_scenarios"] = 1
    engine = initialize_engine(default_config)

    engine.data_manager = ScenarioDataManager()
    map = ScenarioMap(map_index=0)
    map.attach_to_world()
    engine.enableMouse()
    map.road_network.show_bounding_box(engine)

    # argoverse data set is as the same coordinates as panda3d
    pos = map.get_center_point()
    engine.main_camera.set_bird_view_pos(pos)
    while True:
        map.engine.step()



import time

from metadrive import MetaDriveEnv

if __name__ == '__main__':
    env = MetaDriveEnv(dict(num_scenarios=1000, traffic_density=0.1, start_seed=5000))
    obs, _ = env.reset()
    start = time.time()
    vc = []
    for s in range(1000):
        env.reset(seed=s + 5000)
        print("We have {} vehicles in seed {} map!".format(len(env.engine.traffic_manager.vehicles), s))
        vc.append(len(env.engine.traffic_manager.vehicles))
        if (s + 1) % 1 == 0:
            print(f"{s + 1} | Time Elapse: {time.time() - start}")
    import numpy as np
    print(np.mean(vc), np.min(vc), np.max(vc), np.std(vc))



import numpy as np

from metadrive.constants import TerminationState
from metadrive.envs.metadrive_env import MetaDriveEnv

info_keys = [
    "cost", "velocity", "steering", "acceleration", "step_reward", TerminationState.CRASH_VEHICLE,
    TerminationState.OUT_OF_ROAD, TerminationState.SUCCESS
]


def _act(env, action):
    assert env.action_space.contains(action)
    obs, reward, terminated, truncated, info = env.step(action)
    assert env.observation_space.contains(obs)
    assert np.isscalar(reward)
    assert isinstance(info, dict)
    for k in info_keys:
        assert k in info


def test_metadrive_env_rgb():
    env = MetaDriveEnv(dict(image_observation=True))
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        _act(env, env.action_space.sample())
        for x in [-1, 0, 1]:
            env.reset()
            for y in [-1, 0, 1]:
                _act(env, [x, y])
    finally:
        env.close()


if __name__ == '__main__':
    test_metadrive_env_rgb()



import time

from metadrive.envs.top_down_env import TopDownMetaDriveEnvV2

if __name__ == '__main__':
    env = TopDownMetaDriveEnvV2(dict(num_scenarios=10))
    o, _ = env.reset()
    start = time.time()
    action = [0.0, 0.1]
    # print(o.shape)
    for s in range(10000):
        o, r, tm, tc, i = env.step(action)
        if tm or tc:
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "(TopDownEnv) Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    # print(f"(TopDownEnv) Total Time Elapse: {time.time() - start}")



import logging
import math

import cv2
import numpy as np
from metadrive.base_class.base_runnable import BaseRunnable
from metadrive.constants import MapTerrainSemanticColor, MetaDriveType, DrivableAreaProperty
from metadrive.engine.engine_utils import get_global_config
from shapely.geometry import Polygon, MultiPolygon

logger = logging.getLogger(__name__)


class BaseMap(BaseRunnable):
    """
    Base class for Map generation!
    """
    # only used to save and read maps
    FILE_SUFFIX = ".json"

    # define string in json and config
    SEED = "seed"
    LANE_WIDTH = "lane_width"
    LANE_WIDTH_RAND_RANGE = "lane_width_rand_range"
    LANE_NUM = "lane_num"
    BLOCK_ID = "id"
    BLOCK_SEQUENCE = "block_sequence"
    PRE_BLOCK_SOCKET_INDEX = "pre_block_socket_index"

    # generate_method
    GENERATE_CONFIG = "config"
    GENERATE_TYPE = "type"

    # default lane parameter
    MAX_LANE_WIDTH = 4.5
    MIN_LANE_WIDTH = 3.0
    MAX_LANE_NUM = 3
    MIN_LANE_NUM = 2

    def __init__(self, map_config: dict = None, random_seed=None):
        """
        Map can be stored and recover to save time when we access the map encountered before
        """
        assert random_seed is None
        # assert random_seed == map_config[
        #     self.SEED
        # ], "Global seed {} should equal to seed in map config {}".format(random_seed, map_config[self.SEED])
        super(BaseMap, self).__init__(config=map_config)
        self.film_size = (get_global_config()["draw_map_resolution"], get_global_config()["draw_map_resolution"])
        self.road_network = self.road_network_type()

        # A flatten representation of blocks, might cause chaos in city-level generation.
        self.blocks = []

        # Generate map and insert blocks
        # self.engine = get_engine()
        self._generate()
        assert self.blocks, "The generate methods does not fill blocks!"

        #  a trick to optimize performance
        self.spawn_roads = None
        self.detach_from_world()

        # save a backup
        self._semantic_map = None
        self._height_map = None

        if self.engine.global_config["show_coordinates"]:
            self.show_coordinates()

    def _generate(self):
        """Key function! Please overwrite it! This func aims at fill the self.road_network adn self.blocks"""
        raise NotImplementedError("Please use child class like PGMap to replace Map!")

    def attach_to_world(self, parent_np=None, physics_world=None):
        parent_node_path, physics_world = self.engine.worldNP or parent_np, self.engine.physics_world or physics_world
        for block in self.blocks:
            block.attach_to_world(parent_node_path, physics_world)

    def detach_from_world(self, physics_world=None):
        for block in self.blocks:
            block.detach_from_world(self.engine.physics_world or physics_world)

    def get_meta_data(self):
        """
        Save the generated map to map file
        """
        return dict(map_type=self.class_name, map_features=self.get_map_features())

    @property
    def num_blocks(self):
        return len(self.blocks)

    def destroy(self):
        self.detach_from_world()
        if self._semantic_map is not None:
            del self._semantic_map
            self._semantic_map = None
        if self._height_map is not None:
            del self._height_map
            self._height_map = None

        for block in self.blocks:
            block.destroy()
        self.blocks = []

        if self.road_network is not None:
            self.road_network.destroy()
        self.road_network = None

        self.spawn_roads = None

        super(BaseMap, self).destroy()

    @property
    def road_network_type(self):
        raise NotImplementedError

    def get_center_point(self):
        x_min, x_max, y_min, y_max = self.road_network.get_bounding_box()
        return (x_max + x_min) / 2, (y_max + y_min) / 2

    def __del__(self):
        # self.destroy()
        logger.debug("{} 2is being deleted.".format(type(self)))

    def show_coordinates(self):
        pass

    def get_map_features(self, interval=2):
        map_features = self.road_network.get_map_features(interval)

        boundary_line_vector = self.get_boundary_line_vector(interval)

        map_features.update(boundary_line_vector)

        return map_features

    def get_boundary_line_vector(self, interval):
        return {}

    # @time_me
    def get_semantic_map(
        self,
        size=512,
        pixels_per_meter=8,
        color_setting=MapTerrainSemanticColor,
        line_sample_interval=2,
        polyline_thickness=1,
        layer=("lane_line", "lane")
    ):
        """
        Get semantics of the map
        :param size: [m] length and width
        :param pixels_per_meter: the returned map will be in (size*pixels_per_meter * size*pixels_per_meter) size
        :param color_setting: color palette for different attribute. When generating terrain, make sure using
        :param line_sample_interval: [m] It determines the resolution of sampled points.
        :param layer: layer to get
        MapTerrainAttribute
        :return: heightfield image
        """
        if self._semantic_map is None:
            all_lanes = self.get_map_features(interval=line_sample_interval)
            polygons = []
            polylines = []

            points_to_skip = math.floor(DrivableAreaProperty.STRIPE_LENGTH * 2 / line_sample_interval)
            for obj in all_lanes.values():
                if MetaDriveType.is_lane(obj["type"]) and "lane" in layer:
                    polygons.append((obj["polygon"], MapTerrainSemanticColor.get_color(obj["type"])))
                elif "lane_line" in layer and (MetaDriveType.is_road_line(obj["type"])
                                               or MetaDriveType.is_sidewalk(obj["type"])):
                    if MetaDriveType.is_broken_line(obj["type"]):
                        for index in range(0, len(obj["polyline"]) - 1, points_to_skip * 2):
                            if index + points_to_skip < len(obj["polyline"]):
                                polylines.append(
                                    (
                                        [obj["polyline"][index], obj["polyline"][index + points_to_skip]],
                                        MapTerrainSemanticColor.get_color(obj["type"])
                                    )
                                )
                    else:
                        polylines.append((obj["polyline"], MapTerrainSemanticColor.get_color(obj["type"])))

            size = int(size * pixels_per_meter)
            mask = np.zeros([size, size, 4], dtype=np.float32)
            mask[..., 0:] = color_setting.get_color(MetaDriveType.GROUND)
            # create an example bounding box polygon
            # for idx in range(len(polygons)):
            center_p = self.get_center_point()
            for polygon, color in polygons:
                points = [
                    [
                        int((x - center_p[0]) * pixels_per_meter + size / 2),
                        int((y - center_p[1]) * pixels_per_meter) + size / 2
                    ] for x, y in polygon
                ]
                cv2.fillPoly(mask, np.array([points]).astype(np.int32), color=color)
            for line, color in polylines:
                points = [
                    [
                        int((p[0] - center_p[0]) * pixels_per_meter + size / 2),
                        int((p[1] - center_p[1]) * pixels_per_meter) + size / 2
                    ] for p in line
                ]
                cv2.polylines(mask, np.array([points]).astype(np.int32), False, color, polyline_thickness)
            self._semantic_map = mask
        return self._semantic_map

    def get_height_map(
        self,
        size=2048,
        pixels_per_meter=1,
        extension=2,
        height=1,
    ):
        """
        Get height of the map
        :param size: [m] length and width
        :param pixels_per_meter: the returned map will be in (size*pixels_per_meter * size*pixels_per_meter) size
        :param extension: If > 1, the returned height map's drivable region will be enlarged.
        :param height: height of drivable area.
        :return: heightfield image in uint 16 nparray
        """
        if self._height_map is None:
            extension = max(1, extension)
            all_lanes = self.get_map_features()
            polygons = []

            for obj in all_lanes.values():
                if MetaDriveType.is_lane(obj["type"]):
                    polygons.append(obj["polygon"])

            size = int(size * pixels_per_meter)
            mask = np.zeros([size, size, 1])

            center_p = self.get_center_point()
            need_scale = abs(extension - 1) > 1e-1
            for polygon in polygons:
                if need_scale:
                    scaled_polygon = Polygon(polygon).buffer(extension, join_style=2)
                    if isinstance(scaled_polygon, MultiPolygon):
                        scaled_polygons = scaled_polygon.geoms
                    else:
                        scaled_polygons = [scaled_polygon]
                    for scaled_polygon in scaled_polygons:
                        points = [
                            [
                                int(
                                    (scaled_polygon.exterior.coords.xy[0][index] - center_p[0]) * pixels_per_meter +
                                    size / 2
                                ),
                                int((scaled_polygon.exterior.coords.xy[1][index] - center_p[1]) * pixels_per_meter) +
                                size / 2
                            ] for index in range(len(scaled_polygon.exterior.coords.xy[0]))
                        ]
                else:
                    points = [
                        [
                            int((x - center_p[0]) * pixels_per_meter + size / 2),
                            int((y - center_p[1]) * pixels_per_meter) + size / 2
                        ] for x, y in polygon
                    ]
                cv2.fillPoly(mask, np.asarray([points]).astype(np.int32), color=[height])
            self._height_map = mask
        return self._height_map



import logging

from metadrive.component.algorithm.BIG import BIG
from metadrive.component.road_network.node_road_network import NodeRoadNetwork
from metadrive.tests.vis_block.vis_block_base import TestBlock

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test = TestBlock(True)
    global_network = NodeRoadNetwork()

    big = BIG(2, 5, global_network, test.render, test.world, 1010)

    # Since some change to generate function, specify the block num to the big
    big.block_num = len("CrTRXOS")
    big._block_sequence = "CrTRXOS"
    test.vis_big(big)

    # big.generate(BigGenerateMethod.BLOCK_NUM, 10)
    test.run()



# Author: Epihaius
# Date: 2019-06-21
#
# This module contains a class to generate sphere primitives.

from .base import *


class SphereMaker(ModelMaker):
    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, pos):
        self._center = pos

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, radius):
        self._radius = radius

    @property
    def smooth(self):
        return self._smooth

    @smooth.setter
    def smooth(self, smooth):
        self._smooth = smooth

    @property
    def bottom_clip(self):
        return self._bottom_clip

    @bottom_clip.setter
    def bottom_clip(self, clip_height):
        self._bottom_clip = clip_height

    @property
    def top_clip(self):
        return self._top_clip

    @top_clip.setter
    def top_clip(self, clip_height):
        self._top_clip = clip_height

    @property
    def slice(self):
        return self._slice

    @slice.setter
    def slice(self, angle):
        self._slice = angle

    @property
    def thickness(self):
        return self._thickness

    @thickness.setter
    def thickness(self, thickness):
        self._thickness = thickness

    def __init__(
        self,
        center=None,
        radius=1.,
        segments=None,
        smooth=True,
        bottom_clip=-1.,
        top_clip=1.,
        slice=0.,
        thickness=None,
        inverted=False,
        vertex_color=None,
        has_uvs=True,
        tex_units=None,
        tex_offset=None,
        tex_rotation=None,
        tex_scale=None
    ):
        """
        This class generates sphere model primitives with the given parameters:

            center (sequence or None):
                the position of the center of the geometry in object space;
                default is at the origin (the object-space (0., 0., 0.) point);
            radius (float):
                the radius of the sphere; cannot be negative;
                default = 1.;
            segments (dict of ints):
                the number of surface subdivisions:
                    "horizontal":
                        subdivisions along horizontal circles (or circle arcs, if
                        sliced);
                        minimum = 3, default = 20;
                    "vertical":
                        subdivisions along vertical semicircles (or shorter circle
                        arcs, if clipped);
                        minimum = 2, default = 10;
                    "bottom_cap":
                        radial subdivisions of the bottom clipping cap;
                        minimum = 0 (no cap), default = 1;
                    "top_cap":
                        radial subdivisions of the top clipping cap;
                        minimum = 0 (no cap), default = 1;
                    "slice_caps":
                        radial subdivisions of the slice caps;
                        minimum = 0 (no caps), default = 1;
            smooth (bool):
                whether the surface of the mantle should appear smooth or faceted;
                default is True;
            slice (float):
                the angle of the pie slice removed from the sphere, in degrees;
                must be in [0., 360.] range; starts at the positive X-axis;
                default = 0. (no slice);
            bottom_clip (float):
                relative height of the plane that cuts off a bottom part of the sphere;
                must be in [-1., 1.] range;
                default = -1. (no clipping);
            top_clip (float):
                relative height of the plane that cuts off a top part of the sphere;
                must be in [bottom_clip, 1.] range;
                default = 1. (no clipping);
            thickness (float or None):
                offset of inner sphere vertices; cannot be negative;
                default = None (no inner sphere).

        The parameters common to all primitive types are documented in the
        ModelMaker base class.

        The surfaces of this primitive type can be referred to in the relevant
        parameters and properties as follows:
            "main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap",
            "inner_main", "inner_bottom_cap", "inner_top_cap".

        Ranges of vertex indices of each surface can be retrieved through the
        vertex_ranges property, as a dict of (start_index, end_index) tuples
        (empty tuple if the surface was not created), with end_index not
        included in the range.

        """

        surface_ids = (
            "main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap", "inner_main", "inner_bottom_cap",
            "inner_top_cap"
        )

        ModelMaker.__init__(
            self, segments, inverted, vertex_color, has_uvs, tex_units, tex_offset, tex_rotation, tex_scale, surface_ids
        )

        self._center = center
        self._radius = radius
        self._smooth = smooth
        self._bottom_clip = bottom_clip
        self._top_clip = top_clip
        self._slice = slice
        self._thickness = thickness

    def reset(self):

        ModelMaker.reset(self)

        self._center = None
        self._radius = 1.
        self._smooth = True
        self._bottom_clip = -1.
        self._top_clip = 1.
        self._slice = 0.
        self._thickness = None

    def generate(self):

        center = (0., 0., 0.) if self._center is None else self._center
        radius = max(.001, self._radius)
        segs = {} if self._segments is None else self._segments
        segs_h = max(3, segs.get("horizontal", 20))
        segs_v = max(2, segs.get("vertical", 10))
        segs_bc = max(0, segs.get("bottom_cap", 1))
        segs_tc = max(0, segs.get("top_cap", 1))
        segs_sc = max(0, segs.get("slice_caps", 1))
        smooth = self._smooth
        bottom_clip = max(-1., min(1., self._bottom_clip))
        bottom_height = radius * bottom_clip
        top_clip = max(bottom_clip, min(1., self._top_clip))
        top_height = radius * top_clip
        slice = max(0., min(360., self._slice))
        slice_radians = pi * slice / 180.
        delta_angle_h = pi * ((360. - slice) / 180.) / segs_h
        bottom_angle = pi - acos(bottom_height / radius)
        top_angle = acos(top_height / radius)
        delta_angle_v = (pi - bottom_angle - top_angle) / segs_v
        thickness = radius if self._thickness is None else max(0., self._thickness)
        inner_radius = radius - thickness if (top_height - bottom_height) * .5 > thickness else 0.

        if inner_radius:
            inner_bottom_height = bottom_height + thickness
            inner_bottom_clip = inner_bottom_height / inner_radius
            inner_bottom_angle = pi - acos(inner_bottom_height / inner_radius)
            inner_top_height = top_height - thickness
            inner_top_clip = inner_top_height / inner_radius
            inner_top_angle = acos(inner_top_height / inner_radius)
            inner_delta_angle_v = (pi - inner_bottom_angle - inner_top_angle) / segs_v

        inverted = self._inverted
        has_uvs = self._has_uvs
        tex_units = self._tex_units
        tex_offset = self._tex_offset
        tex_rotation = self._tex_rotation
        tex_scale = self._tex_scale
        self._vert_ranges = vert_ranges = {
            "main": (),
            "bottom_cap": (),
            "top_cap": (),
            "slice_start_cap": (),
            "slice_end_cap": (),
            "inner_main": (),
            "inner_bottom_cap": (),
            "inner_top_cap": ()
        }
        stride = 8 if has_uvs else 6  # number of floats on each vertex data row
        values = array.array("f", [])
        indices = array.array("I", [])
        verts = []
        index_offset = 0

        if has_uvs:

            if tex_units and "main" in tex_units:
                tex_size_main = tex_units["main"]
                arc = (2 * pi - slice_radians) * radius
            else:
                tex_size_main = None

            mat_main = self._get_tex_xform("main")

        if bottom_clip > -1.:

            z = bottom_height
            radius_h = sqrt(radius * radius - z * z)

            if has_uvs:

                if tex_units and "bottom_cap" in tex_units:
                    tex_size = tex_units["bottom_cap"]
                else:
                    tex_size = None

                mat = self._get_tex_xform("bottom_cap")

            if segs_bc and -radius < z < radius:

                # Define the bottom cap triangle vertices

                normal = (0., 0., 1. if inverted else -1.)
                u = v = .5

                if has_uvs and mat:
                    u, v = mat.xform_point(Point2(u, v))

                vert = {"pos": (0., 0., z), "normal": normal, "uv": (u, v)}
                verts.append(vert)
                index_offset += segs_h + 2
                r = radius_h / segs_bc

                for i in range(segs_h + 1):

                    angle_h = delta_angle_h * i + (0. if inverted else slice_radians)
                    c = cos(angle_h)
                    s = sin(angle_h) * (-1. if inverted else 1.)
                    x = r * c
                    y = r * s

                    if has_uvs:

                        u = .5 + .5 * c / segs_bc
                        v = .5 + .5 * s * (1. if inverted else -1.) / segs_bc

                        if tex_size:
                            u = (u - .5) * 2. * radius_h / tex_size[0] + .5
                            v = (v - .5) * 2. * radius_h / tex_size[1] + .5

                        if mat:
                            u, v = mat.xform_point(Point2(u, v))

                    else:

                        u = v = 0.

                    vert = {"pos": (x, y, z), "normal": normal, "uv": (u, v)}
                    verts.append(vert)

                # Define the vertex order of the bottom cap triangles

                for i in range(1, segs_h + 1):
                    indices.extend((0, i + 1, i))

                # Define the bottom cap quad vertices

                for i in range(1, segs_bc):

                    r = radius_h * (i + 1) / segs_bc

                    for j in range(segs_h + 1):

                        angle_h = delta_angle_h * j + (0. if inverted else slice_radians)
                        c = cos(angle_h)
                        s = sin(angle_h) * (-1. if inverted else 1.)
                        x = r * c
                        y = r * s

                        if has_uvs:

                            r_ = (i + 1) / segs_bc
                            u = .5 + .5 * c * r_
                            v = .5 + .5 * s * (1. if inverted else -1.) * r_

                            if tex_size:
                                u = (u - .5) * 2. * radius_h / tex_size[0] + .5
                                v = (v - .5) * 2. * radius_h / tex_size[1] + .5

                            if mat:
                                u, v = mat.xform_point(Point2(u, v))

                        else:

                            u = v = 0.

                        vert = {"pos": (x, y, z), "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                # Define the vertex order of the bottom cap quads

                for i in range(1, segs_bc):

                    for j in range(segs_h):
                        vi1 = index_offset + j
                        vi2 = vi1 - segs_h - 1
                        vi3 = vi2 + 1
                        vi4 = vi1 + 1
                        indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi2, vi3))
                        indices.extend((vi2, vi3, vi4) if inverted else (vi1, vi3, vi4))

                    index_offset += segs_h + 1

                vert_ranges["bottom_cap"] = (0, len(verts))

            main_start_index = len(verts)

            # Define the bottom edge vertices

            if has_uvs:
                v = (pi - acos(z / radius)) / pi
                if tex_size_main:
                    v *= pi * radius / tex_size_main[1]
                v_start = v
            else:
                v = 0.

            for i in range(segs_h + 1):

                angle_h = delta_angle_h * i + (0. if inverted else slice_radians)
                x = radius_h * cos(angle_h)
                y = radius_h * sin(angle_h) * (-1. if inverted else 1.)

                if smooth:
                    normal = Vec3(x, y, z).normalized() * (-1. if inverted else 1.)

                if has_uvs:
                    u = i / segs_h
                    if tex_size_main:
                        u *= arc / tex_size_main[0]
                    if mat_main:
                        u, v = mat_main.xform_point(Point2(u, v_start))
                else:
                    u = 0.

                vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

                if not smooth and 0 < i < segs_h:
                    verts.append(vert.copy())

        else:

            main_start_index = 0

            # Define the bottom pole triangle vertices

            if smooth:
                normal = (0., 0., 1. if inverted else -1.)

            for i in range(segs_h):

                if has_uvs:
                    u = i / segs_h
                    v = 0.
                    if tex_size_main:
                        u *= arc / tex_size_main[0]
                    if mat_main:
                        u, v = mat_main.xform_point(Point2(u, v))
                else:
                    u = v = 0.

                vert = {"pos": (0., 0., -radius), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

        vertex_count = len(verts)
        angle_v = bottom_angle + delta_angle_v
        z = radius * -cos(angle_v)

        # Define the vertices along the bottom pole or cap

        radius_h = radius * sin(angle_v)

        if has_uvs:
            v = angle_v / pi
            if tex_size_main:
                v *= pi * radius / tex_size_main[1]
            v_start = v
        else:
            v = 0.

        for i in range(segs_h + 1):

            angle_h = delta_angle_h * i + (0. if inverted else slice_radians)
            x = radius_h * cos(angle_h)
            y = radius_h * sin(angle_h) * (-1. if inverted else 1.)

            if smooth:
                normal = Vec3(x, y, z).normalized() * (-1. if inverted else 1.)

            if has_uvs:
                u = i / segs_h
                if tex_size_main:
                    u *= arc / tex_size_main[0]
                if mat_main:
                    u, v = mat_main.xform_point(Point2(u, v_start))
            else:
                u = 0.

            vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
            verts.append(vert)

            if not smooth and 0 < i < segs_h:
                verts.append(vert.copy())

        # Define the vertex order of the polygons along the bottom pole or cap

        if bottom_clip > -1.:

            n = segs_h if smooth else segs_h * 2 - 1
            f = 1 if smooth else 2

            for i in range(0, segs_h * f, f):

                vi1 = i + index_offset
                vi2 = vi1 + 1
                vi3 = vi2 + n
                vi4 = vi3 + 1
                indices.extend((vi1, vi4, vi3) if inverted else (vi1, vi2, vi3))
                indices.extend((vi1, vi2, vi4) if inverted else (vi2, vi4, vi3))

                if not smooth:
                    self._make_flat_shaded((vi1, vi2, vi3, vi4), verts)

            if smooth:
                index_offset += segs_h + 1

        else:

            for i in range(segs_h):

                j = i + index_offset
                n = 0 if smooth else i
                inds = (j, j + segs_h + 1 + n, j + segs_h + n)
                indices.extend(inds)

                if not smooth:
                    self._make_flat_shaded(inds, verts)

            if smooth:
                index_offset += segs_h

        # Define the main quad vertices

        if not smooth:
            index_offset = len(verts)
            vert_index = len(verts) + segs_h + 1

        n = segs_h + 1 if smooth else segs_h * 2
        f = 1 if smooth else 2

        for i in range(1 if smooth else 0, segs_v - 1):

            angle_v = bottom_angle + delta_angle_v * (i + 1)
            z = radius * -cos(angle_v)
            radius_h = radius * sin(angle_v)

            if has_uvs:
                v = angle_v / pi
                if tex_size_main:
                    v *= pi * radius / tex_size_main[1]
                v_start = v
            else:
                v = 0.

            for j in range(segs_h + 1):

                angle_h = delta_angle_h * j + (0. if inverted else slice_radians)
                x = radius_h * cos(angle_h)
                y = radius_h * sin(angle_h) * (-1. if inverted else 1.)

                if smooth:
                    normal = Vec3(x, y, z).normalized() * (-1. if inverted else 1.)

                if has_uvs:
                    u = j / segs_h
                    if tex_size_main:
                        u *= arc / tex_size_main[0]
                    if mat_main:
                        u, v = mat_main.xform_point(Point2(u, v_start))
                else:
                    u = 0.

                vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

                if not smooth and 0 < j < segs_h:
                    verts.append(vert.copy())

            # Define the vertex order of the main quads

            if i > 0:

                for j in range(0, segs_h * f, f):

                    vi1 = i * n + j + index_offset
                    vi2 = vi1 - n
                    vi3 = vi2 + 1
                    vi4 = vi1 + 1
                    indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi2, vi3))
                    indices.extend((vi2, vi3, vi4) if inverted else (vi1, vi3, vi4))

                    if not smooth:
                        self._make_flat_shaded((vi1, vi2, vi3, vi4), verts)

            if not smooth and i > 0:
                # duplicate latest added vertices
                verts.extend(v.copy() for v in verts[-segs_h * 2:])
                index_offset += segs_h * 2

        if top_clip < 1.:

            # Define the top edge vertices

            z = top_height
            radius_h = sqrt(radius * radius - z * z)

            if has_uvs:
                v = (pi - acos(z / radius)) / pi
                if tex_size_main:
                    v *= pi * radius / tex_size_main[1]
                v_start = v
            else:
                v = 0.

            for i in range(segs_h + 1):

                angle_h = delta_angle_h * i + (0. if inverted else slice_radians)
                x = radius_h * cos(angle_h)
                y = radius_h * sin(angle_h) * (-1. if inverted else 1.)

                if smooth:
                    normal = Vec3(x, y, z).normalized() * (-1. if inverted else 1.)

                if has_uvs:
                    u = i / segs_h
                    if tex_size_main:
                        u *= arc / tex_size_main[0]
                    if mat_main:
                        u, v = mat_main.xform_point(Point2(u, v_start))
                else:
                    u = 0.

                vert = {"pos": (x, y, z), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

                if not smooth and 0 < i < segs_h:
                    verts.append(vert.copy())

        else:

            # Define the top pole triangle vertices

            if smooth:
                normal = (0., 0., -1. if inverted else 1.)

            for i in range(segs_h):

                if has_uvs:
                    u = i / segs_h
                    v = 1.
                    if tex_size_main:
                        u *= arc / tex_size_main[0]
                        v *= pi * radius / tex_size_main[1]
                    if mat_main:
                        u, v = mat_main.xform_point(Point2(u, v))
                else:
                    u = v = 0.

                vert = {"pos": (0., 0., radius), "normal": normal if smooth else None, "uv": (u, v)}
                verts.append(vert)

        index_offset = len(verts) - 1

        # Define the vertex order of the polygons along the top pole or cap

        if top_clip < 1.:

            n = segs_h if smooth else segs_h * 2 - 1
            f = 1 if smooth else 2
            index_offset -= (segs_h - 1) * f + n + 2

            for i in range(0, segs_h * f, f):

                vi1 = i + index_offset
                vi2 = vi1 + 1
                vi3 = vi2 + n
                vi4 = vi3 + 1
                indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi2, vi3))
                indices.extend((vi1, vi4, vi3) if inverted else (vi2, vi4, vi3))

                if not smooth:
                    self._make_flat_shaded((vi1, vi2, vi3, vi4), verts)

            if smooth:
                index_offset += segs_h + 1

        else:

            # Define the vertex order of the top pole triangles

            for i in range(segs_h):

                j = index_offset - i
                n = 0 if smooth else i
                inds = (j, j - segs_h - 1 - n, j - segs_h - n)
                indices.extend(inds)

                if not smooth:
                    self._make_flat_shaded(inds, verts)

        vert_ranges["main"] = (main_start_index, len(verts))

        if top_clip < 1.:

            index_offset = len(verts)

            if segs_tc and -radius < z < radius:

                # Define the top cap triangle vertices

                top_cap_start_index = index_offset

                normal = (0., 0., -1. if inverted else 1.)

                if has_uvs:

                    if tex_units and "top_cap" in tex_units:
                        tex_size = tex_units["top_cap"]
                    else:
                        tex_size = None

                    mat = self._get_tex_xform("top_cap")

                u = v = .5

                if has_uvs and mat:
                    u, v = mat.xform_point(Point2(u, v))

                vert = {"pos": (0., 0., z), "normal": normal, "uv": (u, v)}
                verts.append(vert)
                r = radius_h / segs_tc

                for i in range(segs_h + 1):

                    angle_h = delta_angle_h * i + (0. if inverted else slice_radians)
                    c = cos(angle_h)
                    s = sin(angle_h) * (-1. if inverted else 1.)
                    x = r * c
                    y = r * s

                    if has_uvs:

                        u = .5 + .5 * c / segs_tc
                        v = .5 + .5 * s * (-1. if inverted else 1.) / segs_tc

                        if tex_size:
                            u = (u - .5) * 2. * radius_h / tex_size[0] + .5
                            v = (v - .5) * 2. * radius_h / tex_size[1] + .5

                        if mat:
                            u, v = mat.xform_point(Point2(u, v))

                    else:

                        u = v = 0.

                    vert = {"pos": (x, y, z), "normal": normal, "uv": (u, v)}
                    verts.append(vert)

                # Define the vertex order of the top cap triangles

                for i in range(index_offset + 1, index_offset + segs_h + 1):
                    indices.extend((index_offset, i, i + 1))

                # Define the top cap quad vertices

                for i in range(1, segs_tc):

                    r = radius_h * (i + 1) / segs_tc

                    for j in range(segs_h + 1):

                        angle_h = delta_angle_h * j + (0. if inverted else slice_radians)
                        c = cos(angle_h)
                        s = sin(angle_h) * (-1. if inverted else 1.)
                        x = r * c
                        y = r * s

                        if has_uvs:

                            r_ = (i + 1) / segs_tc
                            u = .5 + .5 * c * r_
                            v = .5 + .5 * s * (-1. if inverted else 1.) * r_

                            if tex_size:
                                u = (u - .5) * 2. * radius_h / tex_size[0] + .5
                                v = (v - .5) * 2. * radius_h / tex_size[1] + .5

                            if mat:
                                u, v = mat.xform_point(Point2(u, v))

                        else:

                            u = v = 0.

                        vert = {"pos": (x, y, z), "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                # Define the vertex order of the top cap quads

                index_offset += segs_h + 2

                for i in range(1, segs_tc):

                    for j in range(segs_h):
                        vi1 = index_offset + j
                        vi2 = vi1 - segs_h - 1
                        vi3 = vi2 + 1
                        vi4 = vi1 + 1
                        indices.extend((vi1, vi3, vi2) if inverted else (vi1, vi4, vi2))
                        indices.extend((vi1, vi4, vi3) if inverted else (vi4, vi3, vi2))

                    index_offset += segs_h + 1

                end = len(values) // stride
                vert_ranges["top_cap"] = (top_cap_start_index, len(verts))

        if segs_sc and slice and bottom_height < radius and top_height > -radius \
                and thickness:

            # Define the slice cap vertices

            for cap_id in ("start", "end"):

                index_offset = slice_cap_start_index = len(verts)

                if cap_id == "start":
                    normal = (0., -1. if inverted else 1., 0.)
                else:
                    angle_h = delta_angle_h * segs_h
                    c_h = cos(angle_h)
                    s_h = -sin(angle_h)
                    normal = Vec3(s_h, -c_h, 0.) * (-1. if inverted else 1.)

                seg_vecs = []

                if inner_radius:

                    inner_pos = []

                    if bottom_clip > -1.:
                        inner_pos.append(Point3(0., 0., inner_bottom_height))
                        seg_vecs.append(Vec3(0., 0., -thickness / segs_sc))

                    for i in range(segs_v + 1):

                        angle_v = bottom_angle + delta_angle_v * i
                        c = -cos(angle_v)
                        r = radius * sin(angle_v)
                        z = radius * c
                        i_angle_v = inner_bottom_angle + inner_delta_angle_v * i
                        i_c = -cos(i_angle_v)
                        i_r = inner_radius * sin(i_angle_v)
                        i_z = inner_radius * i_c

                        if cap_id == "start":
                            p = Point3(r, 0., z)
                            i_p = Point3(i_r, 0., i_z)
                        else:
                            p = Point3(r * c_h, r * s_h, z)
                            i_p = Point3(i_r * c_h, i_r * s_h, i_z)

                        inner_pos.append(i_p)
                        seg_vecs.append((p - i_p) / segs_sc)

                    if top_clip < 1.:
                        inner_pos.append(Point3(0., 0., inner_top_height))
                        seg_vecs.append(Vec3(0., 0., thickness / segs_sc))

                else:

                    z = (top_height + bottom_height) * .5
                    h = (top_height - bottom_height) * .5
                    inner_pos = Point3(0., 0., z)

                    if bottom_clip > -1.:
                        seg_vecs.append(Vec3(0., 0., -h / segs_sc))

                    for i in range(segs_v + 1):

                        angle_v = bottom_angle + delta_angle_v * i
                        c = -cos(angle_v)
                        r = radius * sin(angle_v)
                        z = radius * c

                        if cap_id == "start":
                            p = Point3(r, 0., z)
                        else:
                            p = Point3(r * c_h, r * s_h, z)

                        seg_vecs.append((p - inner_pos) / segs_sc)

                    if top_clip < 1.:
                        seg_vecs.append(Vec3(0., 0., h / segs_sc))

                if has_uvs:

                    cap_name = "slice_{}_cap".format(cap_id)

                    if tex_units and cap_name in tex_units:
                        tex_size = tex_units[cap_name]
                        u_f = 2 * radius / tex_size[0]
                        v_f = 2 * radius / tex_size[1]
                    else:
                        tex_size = None

                    mat = self._get_tex_xform(cap_name)

                if inner_radius:

                    # Define the lower inner central vertex of the slice cap

                    if bottom_clip > -1.:

                        pos = inner_pos[0]

                        if has_uvs:
                            u = .5
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)
                        index_offset += 1

                    # Define the main inner vertices of the slice cap quads

                    for i in range(segs_v + 1):

                        pos = inner_pos[i + (1 if bottom_clip > -1. else 0)]
                        vec = Vec3(pos)

                        if has_uvs:
                            if cap_id == "start":
                                u = .5 + .5 * pos.x / radius * (1. if inverted else -1.)
                            else:
                                vec[2] = 0.
                                u = .5 - .5 * vec.length() / radius * (1. if inverted else -1.)
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                u = (u - .5) * u_f + .5
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                    index_offset += segs_v + 1

                    # Define the upper inner central vertex of the slice cap

                    if top_clip < 1.:

                        pos = inner_pos[-1]

                        if has_uvs:
                            u = .5
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)
                        index_offset += 1

                else:

                    # Define the center vertex of the slice cap

                    if has_uvs:
                        u = .5
                        v = .5 + .5 * inner_pos.z / radius
                        if tex_size:
                            v = (v - .5) * v_f + .5
                        if mat:
                            u, v = mat.xform_point(Point2(u, v))
                    else:
                        u = v = 0.

                    vert = {"pos": inner_pos, "normal": normal, "uv": (u, v)}
                    verts.append(vert)

                for i in range(segs_sc):

                    # Define the lower central vertices of the slice cap

                    if bottom_clip > -1.:

                        i_p = inner_pos[0] if inner_radius else inner_pos
                        pos = i_p + seg_vecs[0] * (i + 1)

                        if has_uvs:
                            u = .5
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                    # Define the main vertices of the slice cap polygons

                    for j in range(segs_v + 1):

                        index = j + (1 if bottom_clip > -1. else 0)
                        i_p = inner_pos[index] if inner_radius else inner_pos
                        pos = i_p + seg_vecs[index] * (i + 1)
                        vec = Vec3(pos)

                        if has_uvs:
                            if cap_id == "start":
                                u = .5 + .5 * pos.x / radius * (1. if inverted else -1.)
                            else:
                                vec[2] = 0.
                                u = .5 - .5 * vec.length() / radius * (1. if inverted else -1.)
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                u = (u - .5) * u_f + .5
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                    # Define the upper central vertices of the slice cap

                    if top_clip < 1.:

                        i_p = inner_pos[-1] if inner_radius else inner_pos
                        pos = i_p + seg_vecs[-1] * (i + 1)

                        if has_uvs:
                            u = .5
                            v = .5 + .5 * pos.z / radius
                            if tex_size:
                                v = (v - .5) * v_f + .5
                            if mat:
                                u, v = mat.xform_point(Point2(u, v))
                        else:
                            u = v = 0.

                        vert = {"pos": pos, "normal": normal, "uv": (u, v)}
                        verts.append(vert)

                    if i == 0 and not inner_radius:

                        # Define the vertex order of the slice cap triangles

                        n = 0 + (1 if bottom_clip > -1. else 0) + (1 if top_clip < 1. else 0)

                        for j in range(segs_v + n):

                            vi1 = index_offset
                            vi2 = vi1 + j + 1
                            vi3 = vi2 + 1

                            if cap_id == "start":
                                indices.extend((vi1, vi2, vi3) if inverted else (vi1, vi3, vi2))
                            else:
                                indices.extend((vi1, vi3, vi2) if inverted else (vi1, vi2, vi3))

                    else:

                        # Define the vertex order of the slice cap quads

                        n = 0 + (1 if bottom_clip > -1. else 0) + (1 if top_clip < 1. else 0)
                        n += 0 if inner_radius else 1

                        for j in range(0 if inner_radius else 1, segs_v + n):

                            vi1 = index_offset + j
                            vi2 = vi1 - segs_v - n - (1 if inner_radius else 0)
                            vi3 = vi2 + 1
                            vi4 = vi1 + 1

                            if cap_id == "start":
                                indices.extend((vi1, vi4, vi2) if inverted else (vi1, vi2, vi4))
                                indices.extend((vi2, vi4, vi3) if inverted else (vi2, vi3, vi4))
                            else:
                                indices.extend((vi1, vi2, vi4) if inverted else (vi1, vi4, vi2))
                                indices.extend((vi2, vi3, vi4) if inverted else (vi2, vi4, vi3))

                    n = 1 + (1 if bottom_clip > -1. else 0) + (1 if top_clip < 1. else 0)
                    index_offset += segs_v + n

                surface_name = "slice_{}_cap".format(cap_id)
                vert_ranges[surface_name] = (slice_cap_start_index, len(verts))

        for vert in verts:

            values.extend(vert["pos"])
            values.extend(vert["normal"])

            if has_uvs:
                values.extend(vert["uv"])

        # Create the geometry structures

        if inner_radius:

            # If a thickness is given, an inner sphere needs to be created to close
            # the surface of the model; its parameters are derived from those of
            # the outer sphere and adjusted to make both spheres fit together.

            segs = {
                "horizontal": segs_h,
                "vertical": segs_v,
                "bottom_cap": segs_bc,
                "top_cap": segs_tc,
                "slice_caps": 0
            }
            surface_names = ("main", "bottom_cap", "top_cap")
            inner_tex_units = {} if tex_units else None

            if tex_units:
                for surface_name in surface_names:
                    if "inner_" + surface_name in tex_units:
                        inner_tex_units[surface_name] = tex_units["inner_" + surface_name]

            inner_tex_offset = {} if tex_offset else None

            if tex_offset:
                for surface_name in surface_names:
                    if "inner_" + surface_name in tex_offset:
                        inner_tex_offset[surface_name] = tex_offset["inner_" + surface_name]

            inner_tex_rot = {} if tex_rotation else None

            if tex_rotation:
                for surface_name in surface_names:
                    if "inner_" + surface_name in tex_rotation:
                        inner_tex_rot[surface_name] = tex_rotation["inner_" + surface_name]

            inner_tex_scale = {} if tex_scale else None

            if tex_scale:
                for surface_name in surface_names:
                    if "inner_" + surface_name in tex_scale:
                        inner_tex_scale[surface_name] = tex_scale["inner_" + surface_name]

            model_maker = SphereMaker(
                None,
                inner_radius,
                segs,
                smooth,
                inner_bottom_clip,
                inner_top_clip,
                slice,
                inverted=not inverted,
                has_uvs=has_uvs,
                tex_units=inner_tex_units,
                tex_offset=inner_tex_offset,
                tex_rotation=inner_tex_rot,
                tex_scale=inner_tex_scale
            )
            node = model_maker.generate()

            # Extend the geometry of the inner sphere with the data of the outer sphere

            geom = node.modify_geom(0)
            vertex_data = geom.modify_vertex_data()
            old_vert_count = vertex_data.get_num_rows()
            old_size = old_vert_count * stride
            vertex_data.set_num_rows(old_vert_count + len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[old_size:] = values

            x, y, z = center

            if x or y or z:
                vertex_data.transform_vertices(Mat4.translate_mat(x, y, z))

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)
                geom.set_vertex_data(vertex_data)

            tris_prim = geom.modify_primitive(0)
            old_row_count = tris_prim.get_num_vertices()
            new_row_count = old_row_count + len(indices)

            if new_row_count < 2**16:
                # make the array compatible with the default index format of the
                # GeomPrimitive (16-bit) if the number of vertices allows it...
                indices = array.array("H", indices)
            else:
                # ...or force the GeomPrimitive to accept more indices by setting
                # its index format to 32-bit
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.set_num_rows(new_row_count)
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[old_row_count:] = indices
            tris_prim.offset_vertices(old_vert_count, old_row_count, new_row_count)

            inner_ranges = model_maker.vertex_ranges

            for surface_name in ("main", "bottom_cap", "top_cap"):

                inner_range = inner_ranges[surface_name]

                if inner_range:
                    vert_ranges["inner_{}".format(surface_name)] = inner_range

            for surface_name in ("main", "bottom_cap", "top_cap", "slice_start_cap", "slice_end_cap"):

                vert_range = vert_ranges[surface_name]

                if vert_range:
                    start, end = vert_range
                    start += old_vert_count
                    end += old_vert_count
                    vert_ranges[surface_name] = (start, end)

        else:

            if has_uvs:
                vertex_format = GeomVertexFormat.get_v3n3t2()
            else:
                vertex_format = GeomVertexFormat.get_v3n3()

            vertex_data = GeomVertexData("sphere_data", vertex_format, Geom.UH_static)
            vertex_data.unclean_set_num_rows(len(verts))
            data_array = vertex_data.modify_array(0)
            memview = memoryview(data_array).cast("B").cast("f")
            memview[:] = values

            x, y, z = center

            if x or y or z:
                vertex_data.transform_vertices(Mat4.translate_mat(x, y, z))

            if self._vertex_color:
                if has_uvs:
                    vertex_format = GeomVertexFormat.get_v3n3c4t2()
                else:
                    vertex_format = GeomVertexFormat.get_v3n3c4()
                vertex_data = vertex_data.set_color(self._vertex_color)

            tris_prim = GeomTriangles(Geom.UH_static)

            if len(indices) < 2**16:
                indices = array.array("H", indices)
            else:
                tris_prim.set_index_type(Geom.NT_uint32)

            tris_array = tris_prim.modify_vertices()
            tris_array.unclean_set_num_rows(len(indices))
            memview = memoryview(tris_array).cast("B").cast(indices.typecode)
            memview[:] = indices

            geom = Geom(vertex_data)
            geom.add_primitive(tris_prim)
            node = GeomNode("sphere")
            node.add_geom(geom)

        return node



from metadrive.constants import TerminationState
from metadrive.envs.marl_envs.marl_inout_roundabout import MultiAgentRoundaboutEnv


def test_respawn():
    out_of_road_cost = 5555
    out_of_road_penalty = 2222
    env = MultiAgentRoundaboutEnv(
        {
            "num_agents": 2,
            "out_of_road_cost": out_of_road_cost,
            "out_of_road_penalty": out_of_road_penalty,
            "delay_done": 0,  # Since we are testing respawn!
            # "use_render": True
            "crash_done": False,
        }
    )
    try:

        assert set(env.action_space.spaces.keys()) == {"agent0", "agent1"}
        assert set(env.config["target_vehicle_configs"].keys()) == {"agent0", "agent1"}
        assert set(env.vehicles.keys()) == set()  # Not initialized yet!

        o, _ = env.reset()

        assert set(env.observations.keys()) == {"agent0", "agent1"}

        assert set(o.keys()) == {"agent0", "agent1"}
        assert set(env.observations.keys()) == {"agent0", "agent1"}
        assert set(env.action_space.spaces.keys()) == {"agent0", "agent1"}
        assert set(env.config["target_vehicle_configs"].keys()) == {"agent0", "agent1"}
        assert set(env.vehicles.keys()) == {"agent0", "agent1"}

        v_id_0 = "agent0"
        v_id_1 = "agent1"
        count = 2
        tracks = []
        done_count = 0
        for i in range(1, 1000):
            o, r, tm, tc, info = env.step({v_id_0: [-1, 1], v_id_1: [1, 1]})
            assert set(o.keys()) == set(r.keys()) == set(info.keys())
            assert set(o.keys()).union({"__all__"}) == set(tm.keys())
            tracks.append(tm)
            if tm[v_id_0]:
                assert info[v_id_0][TerminationState.OUT_OF_ROAD]
                assert info[v_id_0]["cost"] == out_of_road_cost
                assert r[v_id_0] == -out_of_road_penalty
                v_id_0 = "agent{}".format(count)
                count += 1
                done_count += 1
            if tm[v_id_1]:
                assert info[v_id_1][TerminationState.OUT_OF_ROAD]
                assert info[v_id_1]["cost"] == out_of_road_cost
                assert r[v_id_1] == -out_of_road_penalty
                v_id_1 = "agent{}".format(count)
                count += 1
                done_count += 1
            if all(tm.values()):
                raise ValueError()
            if i % 100 == 0:  # Horizon
                v_id_0 = "agent0"
                v_id_1 = "agent1"
                count = 2
                o, _ = env.reset()
                assert set(o.keys()) == {"agent0", "agent1"}
                assert set(env.observations.keys()) == {"agent0", "agent1"}
                assert set(env.action_space.spaces.keys()) == {"agent0", "agent1"}
                assert set(env.config["target_vehicle_configs"].keys()) == {"agent0", "agent1"}
                assert set(env.vehicles.keys()) == {"agent0", "agent1"}
    finally:
        env.close()
    assert done_count > 0
    # print("Finish {} dones.".format(done_count))


def test_delay_done(render=False):
    # Put agent 0 in the left, agent 1 in the right, and let agent 0 dead at first.
    # We wish to see agent 1 hits the dead body of agent 0.
    env = MultiAgentRoundaboutEnv(
        {
            # "use_render": True,
            #
            "target_vehicle_configs": {
                "agent0": {
                    "spawn_longitude": 12,
                    "spawn_lateral": 0,
                    "spawn_lane_index": (">", ">>", 0),
                },
                "agent1": {
                    "spawn_longitude": 10,  # locate a little forward
                    "spawn_lateral": 0,
                    "spawn_lane_index": (">", ">>", 1),
                }
            },
            "num_agents": 2,
            "traffic_density": 0,
            "delay_done": 100,
            "horizon": 100,
            "use_render": render
        }
    )
    try:
        agent0_done = False
        agent1_already_hit = False
        o, _ = env.reset()
        for i in range(1, 300):
            actions = {"agent0": [1, 1], "agent1": [1, 1]}
            if "agent0" not in env.vehicles:
                actions.pop("agent0")
            if "agent1" not in env.vehicles:
                actions.pop("agent1")
            o, r, tm, tc, info = env.step(actions)
            if agent0_done:
                assert "agent0" not in o
                assert "agent0" not in info
                assert "agent0" not in tm
                assert "agent0" not in tc
            if tm.get("agent0") or tc.get("agent0"):
                agent0_done = True
            if agent0_done:
                if info["agent1"][TerminationState.CRASH_VEHICLE]:
                    agent1_already_hit = True
                    # print("Hit!")
            if tm["__all__"]:
                assert agent1_already_hit
                agent0_done = False
                agent1_already_hit = False
                env.reset()
    finally:
        env.close()

    env = MultiAgentRoundaboutEnv({"num_agents": 5, "delay_done": 10, "horizon": 100})
    try:
        env.reset()
        dead = set()
        for _ in range(300):
            o, r, tm, tc, i = env.step({k: [1, 1] for k in env.vehicles.keys()})
            for dead_name in dead:
                assert dead_name not in o
            # print("{} there!".format(env.vehicles.keys()))
            # print("{} dead!".format([kkk for kkk, ddd in d.items() if ddd]))
            for kkk, ddd in tm.items():
                if ddd and kkk != "__all__":
                    dead.add(kkk)
            if tm["__all__"]:
                env.reset()
                dead.clear()
    finally:
        env.close()


if __name__ == '__main__':
    # setup_logger(True)
    test_respawn()
    # test_delay_done(True)



"""

Compares both renders

"""

from __future__ import print_function

from panda3d.core import PNMImage

source_a = "scene.png"
source_b = "scene-rp.png"

write_diff_img = False

img_a = PNMImage(source_a)
img_b = PNMImage(source_b)

w, h = img_a.get_x_size(), img_a.get_y_size()
img_dest = PNMImage(w, h, 3)

error_scale = 10.0

total_diff = 0.0

for x in xrange(w):
    for y in xrange(h):
        val_a = img_a.get_xel(x, y)
        val_b = img_b.get_xel(x, y)

        abs_diff = (val_a - val_b) * error_scale
        r, g, b = abs(abs_diff.x), abs(abs_diff.y), abs(abs_diff.z)

        img_dest.set_xel(x, y, r, g, b)
        total_diff += r + g + b

total_diff /= float(w * h)
total_diff /= error_scale

print("Average difference: ", total_diff, " in RGB: ", total_diff * 255)

img_dest.write("difference.png")



import copy

from metadrive.component.map.pg_map import PGMap
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.intersection import InterSection
from metadrive.component.road_network import Road
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.manager.spawn_manager import SpawnManager
from metadrive.utils import Config

MAIntersectionConfig = dict(
    spawn_roads=[
        Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3),
        -Road(InterSection.node(1, 0, 0), InterSection.node(1, 0, 1)),
        -Road(InterSection.node(1, 1, 0), InterSection.node(1, 1, 1)),
        -Road(InterSection.node(1, 2, 0), InterSection.node(1, 2, 1)),
    ],
    num_agents=30,
    map_config=dict(exit_length=60, lane_num=2),
    top_down_camera_initial_x=80,
    top_down_camera_initial_y=0,
    top_down_camera_initial_z=120
)


class MAIntersectionMap(PGMap):
    def _generate(self):
        length = self.config["exit_length"]

        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"

        # Build a first-block
        last_block = FirstPGBlock(
            self.road_network,
            self.config[self.LANE_WIDTH],
            self.config[self.LANE_NUM],
            parent_node_path,
            physics_world,
            length=length
        )
        self.blocks.append(last_block)

        # Build Intersection
        InterSection.EXIT_PART_LENGTH = length

        if "radius" in self.config and self.config["radius"]:
            extra_kwargs = dict(radius=self.config["radius"])
        else:
            extra_kwargs = {}
        last_block = InterSection(
            1,
            last_block.get_socket(index=0),
            self.road_network,
            random_seed=1,
            ignore_intersection_checking=False,
            **extra_kwargs
        )

        if self.config["lane_num"] > 1:
            # We disable U turn in TinyInter environment!
            last_block.enable_u_turn(True)
        else:
            last_block.enable_u_turn(False)

        last_block.construct_block(parent_node_path, physics_world)
        self.blocks.append(last_block)


class MAIntersectionSpawnManager(SpawnManager):
    def __init__(self, disable_u_turn=False):
        super(MAIntersectionSpawnManager, self).__init__()
        self.disable_u_turn = disable_u_turn

    def update_destination_for(self, agent_id, vehicle_config):
        end_roads = copy.deepcopy(self.engine.global_config["spawn_roads"])
        if self.disable_u_turn:  # Remove the spawn road from end roads
            end_roads = [r for r in end_roads if Road(*vehicle_config["spawn_lane_index"][:2]) != r]
        end_road = -self.np_random.choice(end_roads)  # Use negative road!
        vehicle_config["destination"] = end_road.end_node
        return vehicle_config


class MAIntersectionPGMapManager(PGMapManager):
    def reset(self):
        config = self.engine.global_config
        if len(self.spawned_objects) == 0:
            _map = self.spawn_object(MAIntersectionMap, map_config=config["map_config"], random_seed=None)
        else:
            assert len(self.spawned_objects) == 1, "It is supposed to contain one map in this manager"
            _map = self.spawned_objects.values()[0]
        self.load_map(_map)
        self.current_map.spawn_roads = config["spawn_roads"]


class MultiAgentIntersectionEnv(MultiAgentMetaDrive):
    @staticmethod
    def default_config() -> Config:
        return MultiAgentMetaDrive.default_config().update(MAIntersectionConfig, allow_add_new_key=True)

    def setup_engine(self):
        disable_u_turn = self.config["map_config"]["lane_num"] < 2
        super(MultiAgentIntersectionEnv, self).setup_engine()
        self.engine.update_manager("map_manager", MAIntersectionPGMapManager())
        self.engine.update_manager("spawn_manager", MAIntersectionSpawnManager(disable_u_turn=disable_u_turn))


def _draw():
    env = MultiAgentIntersectionEnv()
    o, _ = env.reset()
    from metadrive.utils.draw_top_down_map import draw_top_down_map
    import matplotlib.pyplot as plt

    plt.imshow(draw_top_down_map(env.current_map))
    plt.show()
    env.close()


def _expert():
    env = MultiAgentIntersectionEnv(
        {
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 240,
                    "num_others": 4,
                    "distance": 50
                },
            },
            "use_AI_protector": True,
            "save_level": 1.,
            "debug_physics_world": True,

            # "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 4,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        tm.update({"total_r": total_r, "episode length": ep_s})
        # env.render(text=d)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis_debug_respawn():
    env = MultiAgentIntersectionEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_physics_world": True,
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": 40,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        action = {k: [0.0, .0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(action)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    env = MultiAgentIntersectionEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            #
            "use_render": True,
            "debug": True,
            "allow_respawn": False,
            "manual_control": True,
            "num_agents": 20,
            "delay_done": 2,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        actions = {k: [0.0, 1.0] for k in env.vehicles.keys()}
        if len(env.vehicles) == 1:
            actions = {k: [-0, 1.0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(actions)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        # render_text = {
        #     "total_r": total_r,
        #     "episode length": ep_s,
        #     "cam_x": env.main_camera.camera_x,
        #     "cam_y": env.main_camera.camera_y,
        #     "cam_z": env.main_camera.top_down_camera_height,
        #     "alive": len(env.vehicles)
        # }
        # env.render(text=render_text)
        # env.render(mode="top_down")
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            env.reset()
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _profile():
    import time
    env = MultiAgentIntersectionEnv({"num_agents": 16})
    obs, _ = env.reset()
    start = time.time()
    for s in range(10000):
        o, r, tm, tc, i = env.step(env.action_space.sample())

        # mask_ratio = env.engine.detector_mask.get_mask_ratio()
        # print("Mask ratio: ", mask_ratio)

        if all(tm.values()):
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    print(f"(MAIntersection) Total Time Elapse: {time.time() - start}")


def _long_run():
    # Please refer to test_ma_Intersection_reward_done_alignment()
    _out_of_road_penalty = 3
    env = MultiAgentIntersectionEnv(
        {
            "num_agents": 32,
            "vehicle_config": {
                "lidar": {
                    "num_others": 8
                }
            },
            **dict(
                out_of_road_penalty=_out_of_road_penalty,
                crash_vehicle_penalty=1.333,
                crash_object_penalty=11,
                crash_vehicle_cost=13,
                crash_object_cost=17,
                out_of_road_cost=19,
            )
        }
    )
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for step in range(10000):
            act = env.action_space.sample()
            o, r, tm, tc, i = env.step(act)
            if step == 0:
                assert not any(tm.values())

            if any(tm.values()):
                print("Current Done: {}\nReward: {}".format(tm, r))
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        print("Info {}: {}\n".format(kkk, i[kkk]))
                print("\n")

            for kkk, rrr in r.items():
                if rrr == -_out_of_road_penalty:
                    assert tm[kkk]

            if (step + 1) % 200 == 0:
                print(
                    "{}/{} Agents: {} {}\nO: {}\nR: {}\nD: {}\nI: {}\n\n".format(
                        step + 1, 10000, len(env.vehicles), list(env.vehicles.keys()),
                        {k: (oo.shape, oo.mean(), oo.min(), oo.max())
                         for k, oo in o.items()}, r, tm, i
                    )
                )
            if tm["__all__"]:
                print('Current step: ', step)
                break
    finally:
        env.close()


def show_map_and_traj():
    import matplotlib.pyplot as plt
    from metadrive.obs.top_down_renderer import draw_top_down_map, draw_top_down_trajectory
    import json
    import cv2
    import pygame
    env = MultiAgentIntersectionEnv()
    env.reset()
    with open("metasvodist_inter_best.json", "r") as f:
        traj = json.load(f)
    m = draw_top_down_map(env.current_map, draw_drivable_area=False, return_surface=True, reverse_color=True)
    m = draw_top_down_trajectory(
        m, traj, entry_differ_color=True, color_list=[(255, 0, 0), (0, 255, 0), (0, 0, 255), (255, 255, 0)]
    )
    ret = cv2.resize(pygame.surfarray.pixels_red(m), (512, 512), interpolation=cv2.INTER_LINEAR)
    #
    plt.imshow(ret)
    plt.show()
    pygame.image.save(m, "image.png")
    env.close()


if __name__ == "__main__":
    # _draw()
    _vis()
    # _vis_debug_respawn()
    # _profiwdle()
    # _long_run()
    # show_map_and_traj()
    # pygame_replay("parking", MultiAgentParkingLotEnv, False, other_traj="metasvodist_parking_best.json")
    # panda_replay(
    #     "parking",
    #     MultiAgentIntersectionEnv,
    #     False,
    #     other_traj="metasvodist_inter.json",
    #     extra_config={
    #         "global_light": True
    #     }
    # )
    # pygame_replay()



import numpy as np


class BicycleModel:
    """
    This model can be used to predict next state
    """
    def __init__(self):
        self.state = dict(x=0, y=0, speed=0, heading_theta=0, velocity_dir=0)

    def reset(self, x, y, speed, heading_theta, velocity_dir):
        """
        heading_theta, velocity_dir in radian
        """
        self.state = dict(x=x, y=y, speed=speed, heading_theta=heading_theta, velocity_dir=velocity_dir)

    def predict(self, dt, control):
        """
        In this model, we formulate the car's dynamic model as Bicycle model
        things need to be finetuned
        mass
        mu*g: a = f/m + \mu * mg
        """
        x = self.state["x"]
        y = self.state["y"]
        v = self.state["speed"]
        phi = self.state["heading_theta"]
        beta = self.state["velocity_dir"]
        pedal, steering = control[0], control[1]
        pedal *= 500
        steering *= 40
        new_beta = np.arctan(0.5 * np.tan(steering / 180 * np.pi))

        if pedal < 0:
            pedal *= 3

        a = pedal / 500 * 3
        af = .5

        new_v = 0
        if v > 1e-5 or a > af:
            new_v = v + (a - af) * dt
            if v * new_v < 0:
                new_v = 0

        new_phi = phi + v / 2 / 2 * np.tan(steering / 180 * np.pi) * dt
        new_x = x + v * np.cos(phi + beta) * dt
        new_y = y + v * np.sin(phi + beta) * dt
        new_state = dict(x=new_x, y=new_y, speed=new_v, heading_theta=new_phi, velocity_dir=new_beta)
        self.state = new_state
        return new_state



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import print_function

import os
import sys
import zipfile
import shutil

from metadrive.render_pipeline.rplibs.six.moves import urllib  # pylint: disable=import-error
from metadrive.render_pipeline.rplibs.six import BytesIO, binary_type


def download_file(url, chunk_size=100 * 1024):
    """ Helper method to download a file displaying a progress bar """
    print("Fetching:", url)
    file_content = None
    progressbar = None

    if sys.version_info.major <= 2:

        # Import progressbar library
        from metadrive.render_pipeline.rplibs.progressbar import FileTransferSpeed, ETA, ProgressBar, Percentage
        from metadrive.render_pipeline.rplibs.progressbar import Bar
        widgets = ['\tDownloading: ', FileTransferSpeed(), ' ', Bar(), Percentage(), '   ', ETA()]
        file_content = []
        bytes_read = 0

        # Progressively download the file
        try:
            usock = urllib.request.urlopen(url)
            file_size = int(usock.headers.get("Content-Length", 1e10))
            print("File size is", round(file_size / (1024**2), 2), "MB")
            progressbar = ProgressBar(widgets=widgets, maxval=file_size).start()
            while True:
                data = usock.read(chunk_size)
                file_content.append(data)
                bytes_read += len(data)
                progressbar.update(bytes_read)
                if not data:
                    break
            usock.close()
        except Exception:
            print("ERROR: Could not fetch", url, "!", file=sys.stderr)
            raise
    else:
        # Don't use progressbar in python 3
        print("Downloading .. (progressbar disabled due to python 3 build)")
        try:
            usock = urllib.request.urlopen(url)
            file_content = []
            while True:
                data = usock.read(chunk_size)
                file_content.append(data)
                if not data:
                    break
            usock.close()
        except Exception:
            print("ERROR: Could not fetch", url, "!", file=sys.stderr)
            raise

    if progressbar:
        progressbar.finish()

    return binary_type().join(file_content)


def download_submodule(author, module_name, dest_path, ignore_list):
    """ Downloads a submodule from the given author and module name, and extracts
    all files which are not on the ignore_list to the dest_path.

    Example: download_submodule("tobspr", "RenderPipeline", ".", ["README.md", "LICENSE"])
    """

    # Make directory, if it does not exist yet
    if not os.path.isdir(dest_path):
        os.makedirs(dest_path)

    # Construct download url
    source_url = "https://github.com/" + author + "/" + module_name + "/archive/master.zip"
    prefix = module_name + "-master"

    # Extract the zip
    zip_ptr = BytesIO(download_file(source_url))
    print("Extracting ZIP ...")

    try:
        zip_handle = zipfile.ZipFile(zip_ptr)
    except zipfile.BadZipfile:
        print("ERROR: Invalid zip file!", file=sys.stderr)
        sys.exit(3)

    if zip_handle.testzip() is not None:
        print("ERROR: Invalid zip file checksums!", file=sys.stderr)
        sys.exit(1)

    num_files, num_dirs = 0, 0

    for fname in zip_handle.namelist():
        rel_name = fname.replace(prefix, "").replace("\\", "/").lstrip("/")
        if not rel_name:
            continue

        is_file = not rel_name.endswith("/")
        rel_name = dest_path.rstrip("/\\") + "/" + rel_name

        # Files
        if is_file:
            for ignore in ignore_list:
                if ignore in rel_name:
                    break
            else:
                with zip_handle.open(fname, "r") as source, open(rel_name, "wb") as dest:
                    shutil.copyfileobj(source, dest)
                num_files += 1

        # Directories
        else:
            if not os.path.isdir(rel_name):
                os.makedirs(rel_name)
            num_dirs += 1

    print("Extracted", num_files, "files and", num_dirs, "directories")



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six import itervalues, iteritems

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.render_target import RenderTarget
from metadrive.render_pipeline.rpcore.loader import RPLoader


class RenderStage(RPObject):
    """ This class is the abstract class for all stages used in the pipeline.
    It represents a part of the pipeline render process. Each stage specifies
    which pipes it uses and which pipes it produces. A pipe can be seen as a
    texture, which gets modified. E.g. the gbuffer pass produces the gbuffer
    pipe, the ambient occlusion pass produces the occlusion pipe and so on. The
    lighting pass can then specify which pipes it needs and compute the image.
    Using a pipe system ensures that new techniques can be inserted easily,
    without the other techniques even being aware of them """

    required_inputs = []
    required_pipes = []

    produced_inputs = {}
    produced_pipes = {}
    produced_defines = {}

    disabled = False

    def __init__(self, pipeline):
        """ Creates a new render stage """
        RPObject.__init__(self)
        self.stage_id = self.__class__.__name__
        self._pipeline = pipeline
        self._active = True
        self._targets = {}

    def create(self):
        """ This method should setup the stage and create the pipes """
        raise NotImplementedError()

    def reload_shaders(self):
        """ This method should set all required shaders, there should be no
        shaders set in the create method, because the shader auto config is not
        generated there """
        pass

    def set_shader_input(self, *args):
        """ This method sets a shader input on all stages, which is mainly used
        by the stage manager """
        for target in itervalues(self._targets):
            target.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        """ This method sets shader inputs on all stages, which is mainly used
        by the stage manager """
        for target in itervalues(self._targets):
            target.set_shader_inputs(**kwargs)

    def update(self):
        """ This method gets called every frame, and can be overridden by render
        stages to perform custom updates """
        pass

    @property
    def active(self):
        """ Returns whether *all* targets of the stage are active """
        return self._active

    @active.setter
    def active(self, state):
        """ Enables or disables this stage. In case the stage is disabled, it will
        not get updated anymore, and all stages are distabled """
        if self._active != state:
            self._active = state
            for target in itervalues(self._targets):
                target.active = self._active

    def create_target(self, name):
        """ Creates a new render target and binds it to this stage """
        # Format the name like Plugin:Stage:Name, so it can be easily
        # found in pstats below the plugin cagetory
        name = self._get_plugin_id() + ":" + self.stage_id + ":" + name
        if name in self._targets:
            return self.error("Overriding existing target: " + name)
        self._targets[name] = RenderTarget(name)
        return self._targets[name]

    def remove_target(self, target):
        """ Removes a previously registered target. This unregisters the
        target, as well as removing it from the list of assigned targets. """
        target.remove()
        target_key = None
        for key, value_target in iteritems(self._targets):
            if target == value_target:
                target_key = key
                break
        del self._targets[target_key]

    def _get_shader_handle(self, path, *args):
        """ Returns a handle to a Shader object, containing all sources passed
        as arguments. The path argument will be used to locate shaders if no
        absolute path is given. This is the internal method used in load_shader
        and load_plugin_shader. """
        assert len(args) > 0 and len(args) <= 3
        path_args = []

        for source in args:
            for prefix in ("/$$rpconfig", "/$$rp/shader", "/$$rptemp"):
                if prefix in source:
                    path_args.append(source)
                    break
            else:
                path_args.append(path.format(source))

        # If only one shader is specified, assume its a postprocess fragment shader,
        # and use the default vertex shader
        if len(args) == 1:
            path_args = ["/$$rp/shader/default_post_process.vert.glsl"] + path_args
        return RPLoader.load_shader(*path_args)

    def _get_plugin_id(self):
        """ Returns the id of the plugin which created this stage. This is done
        by extracting the name of the plugin from the module name """
        if "rpcore.stages" in self.__class__.__module__:
            return "render_pipeline_internal"
        return str(self.__class__.__module__).split(".")[-2]

    def load_shader(self, *args):
        """ Loads a shader from the given args. If only one argument is passed,
        the default template for the stage is loaded. If two arguments are
        passed, the first argument should be the vertex shader and the second
        argument should be the fragment shader. If three arguments are passed,
        the order should be vertex, fragment, geometry """
        return self._get_shader_handle("/$$rp/shader/{0}", *args)

    def load_plugin_shader(self, *args):
        """ Loads a shader from the plugin directory. This method is useful
        for RenderStages created by plugins. For a description of the arguments,
        see the load_shader function. """
        shader_path = "rpplugins/" + self._get_plugin_id() + "/shader/{0}"
        return self._get_shader_handle(shader_path, *args)

    def handle_window_resize(self):
        """ This method gets called when the window gets resized. By default,
        this just resizes all render targets. """
        self.set_dimensions()
        for target in itervalues(self._targets):
            target.consider_resize()

    def set_dimensions(self):
        """ This method should set the dimensions on all targets which don't
        have a relative constraint, and also the size of all images. This
        is called after initialization, and when the window resized. """
        pass



from metadrive.component.opendrive_block.opendrive_block import OpenDriveBlock
from metadrive.component.road_network.edge_road_network import OpenDriveRoadNetwork
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.asset_loader import initialize_asset_loader
from metadrive.tests.vis_block.vis_block_base import TestBlock
from metadrive.utils.opendrive.map_load import load_opendrive_map

if __name__ == "__main__":
    from metadrive.engine.engine_utils import initialize_engine, set_global_random_seed
    from metadrive.envs.base_env import BASE_DEFAULT_CONFIG

    default_config = BASE_DEFAULT_CONFIG
    default_config["use_render"] = True
    default_config["debug"] = True
    default_config["show_coordinates"] = True
    default_config["debug_static_world"] = True
    engine = initialize_engine(default_config)
    set_global_random_seed(0)

    # load map
    initialize_asset_loader(engine)
    map = load_opendrive_map(AssetLoader.file_path("carla", "CARLA_town01.xodr", return_raw_style=False))
    global_network = OpenDriveRoadNetwork()
    i = 0
    blocks = []
    for road in map.roads:
        for section in road.lanes.lane_sections:
            block = OpenDriveBlock(i, global_network, 0, section)
            block.construct_block(engine.render, engine.physics_world)
            blocks.append(block)
            i += 1

    # engine.enableMouse()
    global_network.show_bounding_box(engine, (1, 0, 0, 1))
    lanes = [lane_info.lane for lane_info in global_network.graph.values()]
    engine.show_lane_coordinates(lanes)

    res_x_min, res_x_max, res_y_min, res_y_max = global_network.bounding_box
    engine.camera.setPos((res_x_min + res_x_max) / 2, -(res_y_min + res_y_max) / 2, 700)

    while True:
        engine.step()



__all__ = ['BaseDumper', 'SafeDumper', 'Dumper']

from emitter import *
from serializer import *
from representer import *
from resolver import *


class BaseDumper(Emitter, Serializer, BaseRepresenter, BaseResolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        Representer.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)


class SafeDumper(Emitter, Serializer, SafeRepresenter, Resolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        SafeRepresenter.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)


class Dumper(Emitter, Serializer, Representer, Resolver):
    def __init__(
        self,
        stream,
        default_style=None,
        default_flow_style=None,
        canonical=None,
        indent=None,
        width=None,
        allow_unicode=None,
        line_break=None,
        encoding=None,
        explicit_start=None,
        explicit_end=None,
        version=None,
        tags=None
    ):
        Emitter.__init__(
            self,
            stream,
            canonical=canonical,
            indent=indent,
            width=width,
            allow_unicode=allow_unicode,
            line_break=line_break
        )
        Serializer.__init__(
            self,
            encoding=encoding,
            explicit_start=explicit_start,
            explicit_end=explicit_end,
            version=version,
            tags=tags
        )
        Representer.__init__(self, default_style=default_style, default_flow_style=default_flow_style)
        Resolver.__init__(self)



from metadrive.envs.top_down_env import TopDownMetaDrive

if __name__ == '__main__':
    env = TopDownMetaDrive(
        dict(
            # num_scenarios=100,
            use_render=True,
            # manual_control=True,

            # Extremely easy environment
            num_scenarios=100,
            traffic_density=0.0,
            # map="S",
            start_seed=5000,
            # frame_stack=tune.grid_search([1, 3, 5, 7]),
            # frame_stack=tune.grid_search([1, 3, 5]),
        )
    )
    env.reset()
    try:
        ep_reward = 0
        while True:
            o, r, tm, tc, i = env.step([0.01, 1])
            # print("Obs shape {}, reward {:.8f}, done {}, info {}".format(o.shape, r, d, i))
            env.render()

            ep_reward += r
            if tm or tc:
                # print("Episode reward: ", ep_reward)
                ep_reward = 0
                env.reset()

    finally:
        env.close()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

# Disable the unused variable warning - this occurs since we read a lot more
# properties from the IES file than we actually need
# pylint: disable=unused-variable

from __future__ import print_function

import re

from panda3d.core import PTAFloat, Filename, SamplerState, VirtualFileSystem
from panda3d.core import get_model_path
from direct.stdpy.file import open, join, isfile

from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error

from metadrive.render_pipeline.rpcore.native import IESDataset
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.rpobject import RPObject


class InvalidIESProfileException(Exception):
    """ Exception which is thrown when an error occurs during loading an IES
    Profile """


class IESProfileLoader(RPObject):
    """ Loader class to load .IES files and create an IESDataset from it.
    It generates a LUT for each loaded ies profile which is used by the lighting
    pipeline later on. """

    # Supported IES Profiles
    PROFILES = [
        "IESNA:LM-63-1986", "IESNA:LM-63-1991", "IESNA91", "IESNA:LM-63-1995", "IESNA:LM-63-2002",
        "ERCO Leuchten GmbH  BY: ERCO/LUM650/8701", "ERCO Leuchten GmbH"
    ]

    # Regexp for extracting keywords
    KEYWORD_REGEX = re.compile(r"\[([A-Za-z0-8_-]+)\](.*)")

    def __init__(self, pipeline):
        RPObject.__init__(self)
        self._pipeline = pipeline
        self._entries = []
        self._max_entries = 32
        self._create_storage()

    def _create_storage(self):
        """ Internal method to create the storage for the profile dataset textures """
        self._storage_tex = Image.create_3d("IESDatasets", 512, 512, self._max_entries, "R16")
        self._storage_tex.set_minfilter(SamplerState.FT_linear)
        self._storage_tex.set_magfilter(SamplerState.FT_linear)
        self._storage_tex.set_wrap_u(SamplerState.WM_clamp)
        self._storage_tex.set_wrap_v(SamplerState.WM_repeat)
        self._storage_tex.set_wrap_w(SamplerState.WM_clamp)

        self._pipeline.stage_mgr.inputs["IESDatasetTex"] = self._storage_tex
        self._pipeline.stage_mgr.defines["MAX_IES_PROFILES"] = self._max_entries

    def load(self, filename):
        """ Loads a profile from a given filename and returns the internal
        used index which can be assigned to a light."""

        # Make sure the user can load profiles directly from the ies profile folder
        data_path = join("/$$rp/data/ies_profiles/", filename)
        if isfile(data_path):
            filename = data_path

        # Make filename unique
        fname = Filename.from_os_specific(filename)
        if not VirtualFileSystem.get_global_ptr().resolve_filename(fname, get_model_path().get_value(), "ies"):
            self.error("Could not resolve", filename)
            return -1
        fname = fname.get_fullpath()

        # Check for cache entries
        if fname in self._entries:
            return self._entries.index(fname)

        # Check for out of bounds
        if len(self._entries) >= self._max_entries:
            # TODO: Could remove unused profiles here or regenerate texture
            self.warn("Cannot load IES Profile, too many loaded! (Maximum: 32)")

        # Try loading the dataset, and see what happes
        try:
            dataset = self._load_and_parse_file(fname)
        except InvalidIESProfileException as msg:
            self.warn("Failed to load profile from", filename, ":", msg)
            return -1

        if not dataset:
            return -1

        # Dataset was loaded successfully, now copy it
        dataset.generate_dataset_texture_into(self._storage_tex, len(self._entries))
        self._entries.append(fname)

        return len(self._entries) - 1

    def _load_and_parse_file(self, pth):
        """ Loads a .IES file from a given filename, returns an IESDataset
        which is used by the load function later on. """
        self.debug("Loading ies profile from", pth)

        try:
            with open(pth, "r") as handle:
                lines = handle.readlines()
        except IOError as msg:
            self.error("Failed to open", pth, ":", msg)
            return None

        lines = [i.strip() for i in lines]

        # Parse version header
        self._check_version_header(lines.pop(0))

        # Parse arbitrary amount of keywords
        keywords = self._extract_keywords(lines)  # noqa

        # Next line should be TILT=NONE according to the spec
        if lines.pop(0) != "TILT=NONE":
            raise InvalidIESProfileException("Expected TILT=NONE line, but none found!")

        # From now on, lines do not matter anymore, instead everything is
        # space seperated
        new_parts = (' '.join(lines)).replace(",", " ").split()

        def read_int():
            return int(new_parts.pop(0))

        def read_float():
            return float(new_parts.pop(0))

        # Amount of Lamps
        if read_int() != 1:
            raise InvalidIESProfileException("Only 1 Lamp supported!")

        # Extract various properties
        lumen_per_lamp = read_float()  # noqa
        candela_multiplier = read_float()  # noqa
        num_vertical_angles = read_int()
        num_horizontal_angles = read_int()

        if num_vertical_angles < 1 or num_horizontal_angles < 1:
            raise InvalidIESProfileException("Invalid of vertical/horizontal angles!")

        photometric_type = read_int()  # noqa
        unit_type = read_int()

        # Check for a correct unit type, should be 1 for meters and 2 for feet
        if unit_type not in [1, 2]:
            raise InvalidIESProfileException("Invalid unit type")

        width = read_float()  # noqa
        length = read_float()  # noqa
        height = read_float()  # noqa
        ballast_factor = read_float()  # noqa
        future_use = read_float()  # Unused field for future usage  # noqa
        input_watts = read_float()  # noqa

        # Read vertical angles
        vertical_angles = [read_float() for i in range(num_vertical_angles)]
        horizontal_angles = [read_float() for i in range(num_horizontal_angles)]

        candela_values = []
        candela_scale = 0.0

        for i in range(num_horizontal_angles):
            vertical_data = [read_float() for i in range(num_vertical_angles)]
            candela_scale = max(candela_scale, max(vertical_data))
            candela_values += vertical_data

        # Rescale values, divide by maximum
        candela_values = [i / candela_scale for i in candela_values]

        if len(new_parts) != 0:
            self.warn("Unhandled data at file-end left:", new_parts)

            # Dont abort here, some formats like those from ERCO Leuchten GmbH
            # have an END keyword, just ignore everything after the data was
            # read in.

        dataset = IESDataset()
        dataset.set_vertical_angles(self._list_to_pta(vertical_angles))
        dataset.set_horizontal_angles(self._list_to_pta(horizontal_angles))
        dataset.set_candela_values(self._list_to_pta(candela_values))

        # Testing code to write out the LUT
        # from panda3d.core import Texture
        # tex = Texture("temp")
        # tex.setup_3d_texture(512, 512, 1, Image.T_float, Image.F_r16)
        # dataset.generate_dataset_texture_into(tex, 0)
        # tex.write("generated.png")

        return dataset

    def _list_to_pta(self, list_values):
        """ Converts a list to a PTAFloat """
        pta = PTAFloat.empty_array(len(list_values))
        for i, val in enumerate(list_values):
            pta[i] = val
        return pta

    def _check_version_header(self, first_line):
        """ Checks if the IES version header is correct and the specified IES
        version is supported """
        if first_line not in self.PROFILES:
            raise InvalidIESProfileException("Unsupported Profile: " + first_line)

    def _extract_keywords(self, lines):
        """ Extracts the keywords from a list of lines, and removes all lines
        containing keywords """
        keywords = {}
        while lines:
            line = lines.pop(0)
            if not line.startswith("["):

                # Special format used by some IES files, indicates end of properties
                # By just checking for the tilt keyword instead of validating each line,
                # we can read even malformed lines, like those from ERCO Leuchten GmbH
                if line != "TILT=NONE":
                    continue

                # No keyword, we popped the line already tho, so append it again
                lines.insert(0, line)
                return keywords
            else:

                # Try matching the keywords
                match = self.KEYWORD_REGEX.match(line)
                if match:
                    key, val = match.group(1, 2)
                    keywords[key.strip()] = val.strip()
                else:
                    raise InvalidIESProfileException("Invalid keyword line: " + line)

        return keywords



# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from metadrive.third_party.kitsunetsuki.base.matrices import get_object_matrix

from . import spec


class VertexMixin(object):
    def make_vertex(
        self, obj_matrix, gltf_primitive, mesh, polygon, vertex, vertex_id, loop_id, use_smooth=False, can_merge=False
    ):
        # CO
        co = vertex.co
        if not self._z_up:
            co = self._matrix @ co
        if can_merge and not self._pose_freeze:
            co = obj_matrix @ co

        self._buffer.write(gltf_primitive['attributes']['POSITION'], *tuple(co))

        # normals
        # normal = vertex.normal if use_smooth else polygon.normal
        normal = mesh.loops[loop_id].normal if use_smooth else polygon.normal
        if not self._z_up:
            normal = self._matrix @ normal
        if can_merge and not self._pose_freeze:
            normal = obj_matrix.to_euler().to_matrix() @ normal

        self._buffer.write(gltf_primitive['attributes']['NORMAL'], *tuple(normal))

        # shape keys
        for i, sk_name in enumerate(gltf_primitive['extras']['targetNames']):
            sk_data = mesh.shape_keys.key_blocks[sk_name]
            sk_co = sk_data.data[vertex_id].co
            if not self._z_up:
                sk_co = self._matrix @ sk_co

            self._buffer.write(gltf_primitive['targets'][i]['POSITION'], *tuple(sk_co - co))

    def _write_uv(self, gltf_primitive, uv_id, u, v):
        texcoord = 'TEXCOORD_{}'.format(uv_id)
        if texcoord not in gltf_primitive['attributes']:
            channel = self._buffer.add_channel(
                {
                    'componentType': spec.TYPE_FLOAT,
                    'type': 'VEC2',
                    'extras': {
                        'reference': texcoord,
                    },
                }
            )
            gltf_primitive['attributes'][texcoord] = channel['bufferView']

        self._buffer.write(gltf_primitive['attributes'][texcoord], u, 1 - v)

    def _write_tbs(self, obj_matrix, gltf_primitive, t, b, s, can_merge=False):
        if not self._z_up:
            t = self._matrix @ t
        if can_merge and not self._pose_freeze:
            # t = obj_matrix @ t
            t = obj_matrix.to_euler().to_matrix() @ t
        x, y, z = t

        if 'TANGENT' not in gltf_primitive['attributes']:
            channel = self._buffer.add_channel(
                {
                    'componentType': spec.TYPE_FLOAT,
                    'type': 'VEC4',
                    'extras': {
                        'reference': 'TANGENT',
                    },
                }
            )
            gltf_primitive['attributes']['TANGENT'] = channel['bufferView']

        self._buffer.write(gltf_primitive['attributes']['TANGENT'], x, y, z, s)

    def _write_joints_weights(self, gltf_primitive, joints_num, joints_weights):
        for i, joint_weight in enumerate(joints_weights):
            # prepare joints buffer channel
            joints = 'JOINTS_{}'.format(i)
            if joints not in gltf_primitive['attributes']:
                if joints_num > 255:
                    ctype = spec.TYPE_UNSIGNED_SHORT
                else:
                    ctype = spec.TYPE_UNSIGNED_BYTE

                # Unity glTF importer (UniVRM/UniGLTF) compatibility
                if self._output.endswith('.vrm'):
                    ctype = spec.TYPE_UNSIGNED_SHORT

                channel = self._buffer.add_channel(
                    {
                        'componentType': ctype,
                        'type': 'VEC4',
                        'extras': {
                            'reference': joints,
                        },
                    }
                )
                gltf_primitive['attributes'][joints] = channel['bufferView']

            # write 4 joints
            keys = tuple(zip(*joint_weight))[0]
            assert len(keys) == 4
            self._buffer.write(gltf_primitive['attributes'][joints], *keys)

            # prepare weights buffer channel
            weights = 'WEIGHTS_{}'.format(i)
            if weights not in gltf_primitive['attributes']:
                channel = self._buffer.add_channel(
                    {
                        'componentType': spec.TYPE_FLOAT,
                        'type': 'VEC4',
                        'extras': {
                            'reference': weights,
                        },
                    }
                )
                gltf_primitive['attributes'][weights] = channel['bufferView']

            # write 4 weights
            values = tuple(zip(*joint_weight))[1]
            assert len(values) == 4
            self._buffer.write(gltf_primitive['attributes'][weights], *values)
